"""Module for Fantom-specific utilities and errors. Since Fantom is
Ethereum-compatible, the utilities implementation inherits from the
pantos.common.blockchains.ethereum module.

"""
from pantos.common.blockchains.base import BlockchainUtilitiesError
from pantos.common.blockchains.enums import Blockchain
from pantos.common.blockchains.ethereum import EthereumUtilities
from pantos.common.blockchains.ethereum import EthereumUtilitiesError


class FantomUtilitiesError(EthereumUtilitiesError):
    """Exception class for all Fantom utilities errors.

    """
    pass


class FantomUtilities(EthereumUtilities):
    """Class for Fantom-specific utilities.

    """
    @classmethod
    def get_blockchain(cls) -> Blockchain:
        # Docstring inherited
        return Blockchain.FANTOM

    @classmethod
    def get_error_class(cls) -> type[BlockchainUtilitiesError]:
        # Docstring inherited
        return FantomUtilitiesError



"""Module that defines blockchain-specific enumerations.

"""
import enum


class Blockchain(enum.IntEnum):
    """Enumeration of supported blockchain networks.

    """
    ETHEREUM = 0
    BNB_CHAIN = 1
    # Decommissioned: BITCOIN_RSK = 2
    AVALANCHE = 3
    SOLANA = 4
    POLYGON = 5
    CRONOS = 6
    FANTOM = 7
    CELO = 8
    # Decomissioned: AURORA = 9

    @property
    def name_in_pascal_case(self) -> str:
        """The name of the blockchain network in pascal case.

        """
        return ''.join(word.capitalize() for word in self.name.split('_'))

    @staticmethod
    def from_name(name: str) -> 'Blockchain':
        """Find an enumeration member by its name.

        Parameters
        ----------
        name : str
            The name to search for.

        Raises
        ------
        NameError
            If no enumeration member can be found for the given name.

        """
        name_upper = name.upper()
        for blockchain in Blockchain:
            if name_upper == blockchain.name:
                return blockchain
        raise NameError(name)


class ContractAbi(enum.Enum):
    """Enumeration of supported contract ABIs.

    """
    STANDARD_TOKEN = 0
    PANTOS_TOKEN = 1
    PANTOS_HUB = 2
    PANTOS_FORWARDER = 3

    def get_file_name(self, blockchain: Blockchain) -> str:
        """Get the name of the contract ABI file.

        Parameters
        ----------
        blockchain : Blockchain
            The blockchain to get the contract ABI file name for.

        Returns
        -------
        str
            The name of the contract ABI file.

        """
        return f'{blockchain.name.lower()}_{self.name.lower()}.abi'



"""Common generic exceptions and related classes for Pantos.

"""
import abc
import typing

from pantos.common.blockchains.enums import Blockchain

E = typing.TypeVar('E', bound='BaseError')


class ErrorCreator(abc.ABC, typing.Generic[E]):
    """Base class that helps to properly create Pantos errors for a
    hierachy of related classes (like business-logic interactors or
    blockchain clients/utilities).

    """
    @classmethod
    @abc.abstractmethod
    def get_error_class(cls) -> type[E]:
        """Get the main error class of the subclass in the class
        hierarchy.

        Returns
        -------
        type[E]
            The error class.

        """
        pass  # pragma: no cover

    def _create_error(self, message: typing.Optional[str] = None, *,
                      specialized_error_class: typing.Optional[type[E]] = None,
                      **kwargs: typing.Any) -> E:
        """Create an error that can be catched as both the subclass's
        main error and (if specified) the specialized error (as well as
        their supertypes).

        Parameters
        ----------
        message : str, optional
            An explanation of the error (only to be specified if neither
            the subclass's main error class nor the specialized error
            class provides a default message).
        specialized_error_class : type[E], optional
            The specialized error class.
        **kwargs : dict
            Additional information about the error as keyword arguments.

        Returns
        -------
        E
            The error.

        """
        error_class = self.get_error_class()
        error_classes: tuple[type[E], ...] = ()
        if specialized_error_class is not None:
            error_classes += (specialized_error_class, )
        error_classes += (error_class, )

        class Error(*error_classes):  # type: ignore
            pass

        Error.__name__ = (error_class.__name__ if specialized_error_class
                          is None else specialized_error_class.__name__)
        Error.__qualname__ = Error.__name__
        Error.__module__ = error_class.__module__
        return Error(
            **kwargs) if message is None else Error(message=message, **kwargs)


class BaseError(Exception):
    """Base exception class for all Pantos errors.

    Attributes
    ----------
    details : dict
        Additional information about the error.

    """
    def __init__(self, message: str, **kwargs: typing.Any):
        """Construct an error instance.

        Parameters
        ----------
        message : str
            An explanation of the error.
        **kwargs : dict
            Additional information about the error as keyword arguments.

        """
        super().__init__(message)
        self.details = kwargs

    def __str__(self) -> str:
        string = f'{super().__str__()}'
        if self.details is not None:
            for key, value in self.details.items():
                value = value.name if isinstance(value, Blockchain) else value
                string += f' - {key}: {value}'
        return string


class NotInitializedError(BaseError):
    """Error to be raised if a Pantos resource has not been properly
    initialized when it is requested to be used.

    """
    pass



"""Module for loading and parsing a configuration file.

"""
import errno
import importlib.resources
import logging
import os
import pathlib
import typing

import cerberus  # type: ignore
import dotenv
import pyaml_env  # type: ignore
import yaml

from pantos.common.exceptions import BaseError

_logger = logging.getLogger(__name__)

# Ordered by priority
_CONFIGURATION_PATHS = [
    pathlib.Path.home(),
    pathlib.Path.home() / '.config',
    pathlib.Path('/etc/pantos'),
    pathlib.Path('/etc')
]

if pathlib.Path.cwd() != pathlib.Path('/'):
    _CONFIGURATION_PATHS.insert(0, pathlib.Path.cwd())

if os.environ.get('PANTOS_CONFIG'):
    _logger.info('loading configuration from environment variable '
                 'PANTOS_CONFIG')
    _CONFIGURATION_PATHS.insert(0, pathlib.Path(os.environ['PANTOS_CONFIG']))


class _CustomValidator(cerberus.Validator):
    def _validate_one_not_present(self, other: str, field: str, value: str):
        if (bool(value)) == (bool(self.document.get(other))):
            self._error(field, "only one field can be present: " + other)

    def _normalize_coerce_load_if_file(self, value: str):
        path = pathlib.Path(value)
        try:
            # This method may trigger an exception if the path is not valid
            if path.is_file():
                with open(path, 'r') as file:
                    return file.read()
        except OSError as error:
            if error.errno != errno.ENAMETOOLONG:
                raise error
        return value


class ConfigError(BaseError):
    """Exception class for all configuration errors.

    """
    pass


class Config:
    """Class that loads and parses a configuration file and provides
    dictionary-like access to the configuration values.

    Attributes
    ----------
    default_file_name : str
        The default configuration file name to be used when no explicit
        file path is provided for loading the configuration.

    """
    def __init__(self, default_file_name: str):
        """Initialize a configuration instance.

        Parameters
        ----------
        default_file_name : str
            The default configuration file name to be used when no
            explicit file path is provided for loading the
            configuration.

        """
        self.default_file_name = default_file_name
        self.__config_dict: dict[str, typing.Any] | None = None

    def __getitem__(self, key: str) -> typing.Any:
        """Get the configuration value for a given key.

        Parameters
        ----------
        key : str
            The key to get the configuration value for.

        Returns
        -------
        Any
            The configuration value.

        Raises
        ------
        ConfigError
            If the configuration has not been loaded.

        """
        if self.__config_dict is None:
            raise ConfigError('configuration not yet loaded')
        return self.__config_dict[key]

    def __str__(self) -> str:
        return str(self.__config_dict)

    def is_loaded(self) -> bool:
        """Determine if the configuration has been loaded.

        Returns
        -------
        bool
            True if the configuration has been loaded.

        """
        return self.__config_dict is not None

    def load(self, validation_schema: dict[str, typing.Any],
             file_path: str | None = None) -> None:
        """Load the configuration from a file.

        Parameters
        ----------
        validation_schema : dict
            The Cerberus validation schema used for validating the
            loaded configuration.
        file_path : str, optional
            The path to the configuration file to load. If no file path
            is provided, the configuration is loaded from a default
            configuration file path.

        Raises
        ------
        ConfigError
            If the validation schema is invalid, the configuration file
            cannot be found, or the configuration file is invalid.

        """
        path = self.__find_file(file_path)
        _logger.info(f'loading configuration from file {path}')
        config_dict = self.__parse_file(path)
        # Validate the configuration and add default configuration values
        self.__config_dict = self.__validate(config_dict, validation_schema)

    def __find_file(self, file_path: str | None) -> pathlib.Path:
        if file_path is not None:
            # Use the specified configuration file
            path = pathlib.Path(file_path)
            if not path.is_file():
                raise ConfigError(
                    f'no configuration file found at {file_path}')
            return path
        # Find the configuration file at common locations
        for path in _CONFIGURATION_PATHS:
            config_path = path
            try:
                if config_path.is_dir():
                    config_path = config_path / self.default_file_name
                if config_path.is_file():
                    return config_path
            except OSError:
                # Perhaps the path is not readable
                _logger.warning(f'error while reading: {config_path}',
                                exc_info=True)
        # Package resource
        if importlib.resources.is_resource('pantos', self.default_file_name):
            with importlib.resources.path('pantos',
                                          self.default_file_name) as path:
                return path
        # No configuration file found at common locations
        raise ConfigError('no configuration file found')

    def __parse_file(self, path: pathlib.Path) -> dict[str, typing.Any]:
        # List of potential .env file paths
        env_files = [
            pathlib.Path(path.as_posix() + '.env'),
            pathlib.Path(path.with_suffix('.env').as_posix())
        ]
        if os.environ.get('PANTOS_ENV_FILE'):
            _logger.info('loading env variables from environment defined file '
                         'PANTOS_ENV_FILE')
            env_files.insert(0, pathlib.Path(os.environ['PANTOS_ENV_FILE']))

        # Extend env_files with .env paths from _CONFIGURATION_PATHS
        env_files.extend(
            pathlib.Path(str(p)).with_name(
                str(p.name) + self.default_file_name + '.env')
            for p in _CONFIGURATION_PATHS)
        # Iterate over the potential .env file paths
        for env_file in env_files:
            try:
                if env_file.is_file():
                    dotenv.load_dotenv(env_file)
                    _logger.info(f'loaded .env from file {env_file}')
                    break
            except OSError:
                # Perhaps the path is not readable
                _logger.warning(f'error while reading: {env_file}',
                                exc_info=True)
            except Exception:
                _logger.error(f'unable to load .env file {env_file}',
                              exc_info=True)
        # Parse the YAML code in the configuration file
        try:
            return pyaml_env.parse_config(path.as_posix(), default_value='')
        except yaml.YAMLError as error:
            if hasattr(error, 'problem_mark'):
                line = error.problem_mark.line + 1
                column = error.problem_mark.column + 1
                raise ConfigError('YAML code in configuration file invalid at '
                                  f'line {line} and column {column}')
            else:
                raise ConfigError('YAML code in configuration file invalid')

    def __validate(
            self, config_dict: dict[str, typing.Any],
            validation_schema: dict[str, typing.Any]) -> dict[str, typing.Any]:
        # Create the validator and validate the validation schema
        try:
            validator = _CustomValidator(validation_schema)
        except cerberus.schema.SchemaError as error:
            raise ConfigError(f'validation schema invalid: {error}')
        # Validate the configuration
        if not validator.validate(config_dict):
            raise ConfigError(
                f'configuration file invalid: {validator.errors}')
        # Add default configuration values
        return validator.normalized(config_dict)



"""Factory for Pantos blockchain utilities.

"""
import typing

from pantos.common.blockchains.base import BlockchainUtilities
from pantos.common.blockchains.enums import Blockchain
from pantos.common.exceptions import NotInitializedError

_blockchain_utilities: dict[Blockchain, BlockchainUtilities] = {}
"""Blockchain-specific utilities objects."""

_blockchain_utilities_classes = BlockchainUtilities.find_subclasses()
"""Blockchain-specific utilities classes."""


def initialize_blockchain_utilities(
        blockchain: Blockchain, blockchain_node_urls: list[str],
        fallback_blockchain_node_urls: list[str], average_block_time: int,
        required_transaction_confirmations: int,
        transaction_network_id: typing.Optional[int],
        default_private_key: typing.Optional[tuple[str, str]] = None,
        celery_tasks_enabled: bool = False) -> None:
    """Initialize the utilities for the specified blockchain.

    Parameters
    ----------
    blockchain : Blockchain
        The blockchain to initialize a utilities instance for.
    blockchain_node_urls : str
        The URLs of the blockchain nodes to use for the specified
        blockchain.
    fallback_blockchain_node_urls : list[str]
        The URLs of the fallback nodes to use for communication
        with the blockchain network.
    average_block_time : int
        The average time in seconds required to generate a new block of
        the blockchain.
    required_transaction_confirmations : int
        The number of required confirmations for a transaction to be
        considered included in the blockchain.
    transaction_network_id : int or None
        The unique public (i.e. non-Pantos-specific) blockchain network
        identifier (partly called chain ID) to be used for signing
        transactions (to prevent replay attacks between different
        compatible blockchain networks). It is assumed to be the
        identifier of the main or a test network of the specified
        blockchain.
    default_private_key : tuple of str and str, optional
        The keystore value and password of the default private
        key to be used by the blockchain utilities. (default: None).
    celery_tasks_enabled : bool, optional
        If True, Celery tasks are enabled for enhanced functionalities
        (default: False). This requires a proper Celery environment to
        be set up by the project using the blockchain utilities.

    Raises
    ------
    BlockchainUtilitiesError
        If the blockchain-specific utilities cannot be initialized.

    """
    utilities_class = _blockchain_utilities_classes[blockchain]
    _blockchain_utilities[blockchain] = utilities_class(
        blockchain_node_urls, fallback_blockchain_node_urls,
        average_block_time, required_transaction_confirmations,
        transaction_network_id, default_private_key=default_private_key,
        celery_tasks_enabled=celery_tasks_enabled)


def get_blockchain_utilities(blockchain: Blockchain) -> BlockchainUtilities:
    """Factory for blockchain-specific utilities objects.

    Parameters
    ----------
    blockchain : Blockchain
        The blockchain to get the utilities instance for.

    Returns
    -------
    BlockchainUtilities
        A blockchain utilities instance for the specified blockchain.

    Raises
    ------
    NotInitializedError
        If the utilities have not been initialized for the specified
        blockchain.

    """
    try:
        return _blockchain_utilities[blockchain]
    except KeyError:
        raise NotInitializedError(
            f'{blockchain.name} utilities have not been initialized')



"""Module for Polygon-specific utilities and errors. Since Polygon is
Ethereum-compatible, the utilities implementation inherits from the
pantos.common.blockchains.ethereum module.

"""
from pantos.common.blockchains.base import BlockchainUtilitiesError
from pantos.common.blockchains.enums import Blockchain
from pantos.common.blockchains.ethereum import EthereumUtilities
from pantos.common.blockchains.ethereum import EthereumUtilitiesError


class PolygonUtilitiesError(EthereumUtilitiesError):
    """Exception class for all Polygon utilities errors.

    """
    pass


class PolygonUtilities(EthereumUtilities):
    """Class for Polygon-specific utilities.

    """
    @classmethod
    def get_blockchain(cls) -> Blockchain:
        # Docstring inherited
        return Blockchain.POLYGON

    @classmethod
    def get_error_class(cls) -> type[BlockchainUtilitiesError]:
        # Docstring inherited
        return PolygonUtilitiesError



"""Base module that defines the common class and error interfaces for
all blockchain utilities modules.

"""
import abc
import collections
import copy
import dataclasses
import importlib
import importlib.resources
import inspect
import json
import logging
import math
import pathlib
import pkgutil
import random
import typing
import urllib.parse
import uuid

import semantic_version  # type: ignore

from pantos.common.blockchains.enums import Blockchain
from pantos.common.blockchains.enums import ContractAbi
from pantos.common.entities import TransactionStatus
from pantos.common.exceptions import BaseError
from pantos.common.exceptions import ErrorCreator
from pantos.common.types import BlockchainAddress
from pantos.common.types import ContractFunctionArgs

_BASE_CONTRACT_ABI_PACKAGE = 'pantos.common.blockchains.contracts'

MIN_ADAPTABLE_FEE_INCREASE_FACTOR = 1.101
"""The minimum factor for increasing the adaptable fee per gas in
transaction resubmissions.
"""

GENERAL_RPC_ERROR_MESSAGE = 'unreachable'

H = typing.TypeVar('H', bound='BlockchainHandler')
T = typing.TypeVar('T',
                   bound='BlockchainUtilities.TransactionSubmissionRequest')
N = typing.TypeVar('N')
W = typing.TypeVar('W')

_logger = logging.getLogger(__name__)


class BlockchainUtilitiesError(BaseError):
    """Exception class for all blockchain utilities errors.

    """
    pass


class NodeConnectionError(BaseError):
    """Exception class for all node connection errors.

    """
    pass


class MaxTotalFeePerGasExceededError(BlockchainUtilitiesError):
    """Exception to be raised if the maximum total fee per gas would be
    exceeded for a transaction to be submitted.

    """
    def __init__(self, **kwargs: typing.Any):
        # Docstring inherited
        super().__init__('maximum total fee per gas exceeded', **kwargs)


class TransactionNonceTooLowError(BlockchainUtilitiesError):
    """Exception to be raised if a transaction has been submitted with a
    nonce too low.

    """
    def __init__(self, **kwargs: typing.Any):
        # Docstring inherited
        super().__init__('transaction nonce too low', **kwargs)


class TransactionUnderpricedError(BlockchainUtilitiesError):
    """Exception to be raised if a transaction has been submitted as an
    underpriced transaction.

    """
    def __init__(self, **kwargs: typing.Any):
        # Docstring inherited
        super().__init__('transaction underpriced', **kwargs)


class SingleNodeConnectionError(BlockchainUtilitiesError):
    """Exception to be raised if a blockchain specific node
    connection cannot be established.

    """
    def __init__(self, **kwargs: typing.Any):
        # Docstring inherited
        super().__init__('single node connection error', **kwargs)


class ResultsNotMatchingError(NodeConnectionError):
    """Exception to be raised if the results given by the blockchain
    nodes do not match.

    """
    def __init__(self, **kwargs: typing.Any):
        # Docstring inherited
        super().__init__('results not matching', **kwargs)


class NodeConnections(typing.Generic[N]):
    """Class for managing node connections in a blockchain-agnostic way.

    """
    __node_connections: list[N]
    __transaction_method_names: list[str]

    class Wrapper(typing.Generic[W]):
        """Class for wrapping blockchain interactor objects.

        """
        __is_transaction_function: bool
        __objects: list[typing.Any]
        __transaction_method_names: list[str]

        def __init__(self, objects: list[typing.Any], attr_name: str,
                     transaction_method_names: list[str]):
            """Construct the wrapper instance.

            Parameters
            ----------
            objects : list
                The list of wrapped blockchain interactor objects.
            attr_name : str
                The name of the attribute.
            transaction_method_names : list of str
                The list with method names which are sending
                transactions to the blockchain nodes.

            """
            self.__transaction_method_names = transaction_method_names

            self.__is_transaction_function = (attr_name
                                              in transaction_method_names)

            if self.__is_transaction_function:
                self.__objects = [getattr(objects[0], attr_name)]
            else:
                self.__objects = [
                    getattr(object_, attr_name) for object_ in objects
                ]

        def get(self) -> typing.Any:
            """Get the compared results of wrapped objects. If they
            do not match, then a ResultsNotMatchingError is raised.

            Returns
            -------
            Any
                The matching result of wrapped objects.

            Raises
            ------
            ResultsNotMatchingError
                If the results do not match.

            """
            result = self.__objects[0]
            if not all(result == other_result
                       for other_result in self.__objects[1:]):
                raise ResultsNotMatchingError(
                    **{
                        str(index): result
                        for index, result in enumerate(self.__objects)
                    })
            return result

        def get_minimum_result(self, *args: typing.Any,
                               **kwargs: typing.Any) -> int:
            """Get the minimum result value from the wrapped objects.

            Returns
            -------
            int
                The minimum result of the wrapped objects.

            Raises
            ------
            NodeConnectionError
                If result values are not instances of int.

            """
            self.__check_results_are_instances_of_int()
            return min(self.__objects)

        def get_maximum_result(self, *args: typing.Any,
                               **kwargs: typing.Any) -> int:
            """Get the maximum result value from the wrapped objects.

            Returns
            -------
            int
                The maximum result of the wrapped objects.

            Raises
            ------
            NodeConnectionError
                If result values are not instances of int.

            """
            self.__check_results_are_instances_of_int()
            return max(self.__objects)

        def __check_results_are_instances_of_int(self):
            if not all(isinstance(value, int) for value in self.__objects):
                raise NodeConnectionError('the result values from the wrapped '
                                          'objects must be instances of int')

        def __getattr__(self, attr_name: str) -> 'NodeConnections.Wrapper':
            """Create a new wrapper of the currently wrapped objects
            with the given attribute.

            Parameters
            ----------
            attr_name : str
                The attribute name.

            Returns
            -------
            Wrapper
                Object which wraps the currently wrapped objects
                with the given attribute.

            """
            return NodeConnections.Wrapper(self.__objects, attr_name,
                                           self.__transaction_method_names)

        def __getitem__(self, index: typing.Any) -> 'NodeConnections.Wrapper':
            """Forward the retrieval of the item found at the given
            index to the wrapped objects.

            Parameters
            ----------
            index : int
                The index of the requested item.

            Returns
            -------
            Wrapper
                The wrapper object with the inner wrapped items
                retrieved.

            Raises
            ------
            IndexError
                If the list index is out of range.

            """
            self.__objects = [object_[index] for object_ in self.__objects]
            return self

        def __call__(
                self, *args: typing.Any, **kwargs: typing.Any) -> \
                typing.Union['NodeConnections.Wrapper', typing.Any]:
            """Forward the called method to the wrapped objects. If it
            sends a transaction to the blockchain node, it will be
            forwarded to only one of them, randomly chosen.

            Returns
            -------
            Wrapper or Any
                If the called method is a transaction function, the
                result. Otherwise, the wrapper object
                with the inner wrapped objects called.

            """
            if self.__is_transaction_function:
                random_node_index = random.randint(0, len(self.__objects) - 1)
                return self.__objects[random_node_index](*args, **kwargs)
            self.__objects = [
                object_(*args, **kwargs) for object_ in self.__objects
            ]
            return self

    def __init__(self, transaction_method_names: list[str] = []):
        """Construct an instance.

        Parameters
        ----------
        transaction_method_names : list of str
            The list with method names which are sending transactions
            to the blockchain nodes.

        """
        self.__node_connections = []
        self.__transaction_method_names = transaction_method_names

    def add_node_connection(self, node_connection: N) -> None:
        """Add a node connection to the list of node connections.

        Parameters
        ----------
        node_connection : N
            The node connection to be added.

        """
        self.__node_connections.append(node_connection)

    def get_configured_node_connections(self) -> list[N]:
        """Get the node connections.

        Returns
        -------
        list[N]
            The list of node connections.

        """
        return self.__node_connections

    def __getattr__(self, attr_name: str) -> Wrapper:
        """Creates a wrapper object for the node connections with the
        given attribute name.

        Parameters
        ----------
        attr_name : str
            The name of the attribute.

        Raises
        ------
        NodeConnectionError
            If no node connection was added.

        Returns
        -------
        Wrapper
            Object which wraps the current node connections
            with the given attribute.

        """
        if len(self.__node_connections) == 0:
            raise NodeConnectionError(
                'at least one valid connection must be established')
        return NodeConnections.Wrapper(self.__node_connections, attr_name,
                                       self.__transaction_method_names)


class BlockchainHandler(abc.ABC):
    """Base class for all classes that handle blockchain-specific data
    and operations. These include blockchain clients and utilities.

    """
    @classmethod
    @abc.abstractmethod
    def get_blockchain(cls) -> Blockchain:
        """Get the blockchain the handler is implemented for.

        Returns
        -------
        Blockchain
            The supported blockchain.

        """
        pass  # pragma: no cover

    @classmethod
    def get_blockchain_name(cls) -> str:
        """Get the name of the blockchain the handler is implemented
        for.

        Returns
        -------
        str
            The name of the supported blockchain.

        """
        return cls.get_blockchain().name

    @classmethod
    def find_subclasses(cls: type[H]) -> dict[Blockchain, type[H]]:
        """Find all subclasses of the blockchain handler in the
        handler's package.

        Returns
        -------
        dict
            A dictionary of the blockchain handler's subclasses with
            their supported blockchains as keys.

        """
        package_name = cls.__module__.rpartition('.')[0]
        package_path = str(pathlib.Path(inspect.getfile(cls)).parent)
        # Only subclasses in imported modules can be found
        for module in pkgutil.iter_modules([package_path]):
            full_name = f'{package_name}.{module.name}'
            if full_name != 'pantos.common.blockchains.tasks':
                importlib.import_module(full_name)
        blockchain_handlers = {}
        handler_classes = collections.deque[type[H]](cls.__subclasses__())
        while len(handler_classes) > 0:
            handler_class = handler_classes.pop()
            if not inspect.isabstract(handler_class):
                blockchain = handler_class.get_blockchain()
                blockchain_handlers[blockchain] = handler_class
            handler_classes.extendleft(handler_class.__subclasses__())
        return blockchain_handlers


@dataclasses.dataclass
class VersionedContractAbi:
    """Class which encapsulates the ABI definition of a contract
    with support for versioning.

    Attributes
    ----------
    contract_abi : ContractAbi
        Supported contract ABI.
    version : semantic_version.Version
        The semantic version of the contract ABI.

    """
    contract_abi: ContractAbi
    version: semantic_version.Version


@dataclasses.dataclass
class UnhealthyNode:
    """Entity which encapsulates information about an unhealthy node.

    Attributes
    ----------
    node_domain : str
        The domain of the node's URL.
    status : str
        The status of the node.

    """
    node_domain: str
    status: str


class BlockchainUtilities(BlockchainHandler,
                          ErrorCreator[BlockchainUtilitiesError]):
    """Base class for all blockchain utilities classes.

    Attributes
    ----------
    average_block_time : int
        The average time in seconds required to generate a new block of
        the blockchain.
    required_transaction_confirmations : int
        The number of required confirmations for a transaction to be
        considered included in the blockchain.
    transaction_network_id : int or None
        The unique public (i.e. non-Pantos-specific) blockchain network
        identifier (partly called chain ID) to be used for signing
        transactions (to prevent replay attacks between different
        compatible blockchain networks).

    """
    def __init__(self, blockchain_node_urls: list[str],
                 fallback_blockchain_node_urls: list[str],
                 average_block_time: int,
                 required_transaction_confirmations: int,
                 transaction_network_id: typing.Optional[int],
                 default_private_key: typing.Optional[tuple[str, str]] = None,
                 celery_tasks_enabled: bool = False):
        """Construct a blockchain utilities instance.

        Parameters
        ----------
        blockchain_node_urls : list of str
            The URLs of the nodes to use for communication with the
            blockchain network.
        fallback_blockchain_node_urls : list[str]
            The URLs of the fallback nodes to use for communication
            with the blockchain network.
        average_block_time : int
            The average time in seconds required to generate a new block
            of the blockchain.
        required_transaction_confirmations : int
            The number of required confirmations for a transaction to be
            considered included in the blockchain.
        transaction_network_id : int or None
            The unique public (i.e. non-Pantos-specific) blockchain
            network identifier (partly called chain ID) to be used for
            signing transactions (to prevent replay attacks between
            different compatible blockchain networks). It is assumed to
            be the identifier of the main or a test network of the
            blockchain supported by the blockchain utilities subclass.
        default_private_key : tuple of str and str, optional
            The keystore value and password of the default private
            key to be used by the blockchain utilities. (default: None).
        celery_tasks_enabled : bool, optional
            If True, Celery tasks are enabled for enhanced
            functionalities (default: False). This requires a proper
            Celery environment to be set up by the project using the
            blockchain utilities.

        Raises
        ------
        BlockchainUtilitiesError
            If the blockchain utilities initialization fails.

        """
        if average_block_time <= 0:
            raise self._create_error('average block time must be > 0',
                                     average_block_time=average_block_time)
        if required_transaction_confirmations < 0:
            raise self._create_error(
                'required transaction confirmations must be >= 0',
                required_transaction_confirmations=  # noqa: E251
                required_transaction_confirmations)
        if transaction_network_id is not None and transaction_network_id <= 0:
            raise self._create_error(
                'transaction network ID must be > 0',
                transaction_network_id=transaction_network_id)
        if len(blockchain_node_urls) == 0:
            raise self._create_error(
                'at least one blockchain node URL is expected')
        self.average_block_time = average_block_time
        self.required_transaction_confirmations = \
            required_transaction_confirmations
        self.transaction_network_id = transaction_network_id
        self._blockchain_node_urls = blockchain_node_urls
        self._fallback_blockchain_node_urls = fallback_blockchain_node_urls
        self._default_private_key = (
            None if default_private_key is None else self.decrypt_private_key(
                default_private_key[0], default_private_key[1]))
        self._default_address = (None if self._default_private_key is None else
                                 self.get_address(self._default_private_key))
        self._celery_tasks_enabled = celery_tasks_enabled
        self.__loaded_contract_abis: dict[ContractAbi, list[typing.Any]] = {}

    def create_node_connections(
            self,
            timeout: typing.Optional[typing.Union[float, tuple]] = None) \
            -> NodeConnections:
        """Create blockchain node connections.

        Parameters
        ----------
        timeout : float or tuple
            How long to wait for the server to send data before giving up,
            as a float, or a (connect timeout, read timeout) tuple.

        Returns
        -------
        NodeConnections
            The initialized object with valid connections.

        Raises
        ------
        BlockchainUtilities
            Raised if the node connection cannot be initialized.

        """
        fallback_nodes = self._fallback_blockchain_node_urls.copy()
        node_connections = NodeConnections[typing.Any](
            self._get_transaction_method_names())

        for blockchain_node_url in self._blockchain_node_urls:
            node_connection = self.__create_valid_node_connection(
                blockchain_node_url, fallback_nodes, timeout)
            node_connections.add_node_connection(node_connection)

        return node_connections

    def __create_valid_node_connection(
            self, blockchain_node_url: str,
            fallback_blockchain_node_urls: list[str],
            timeout: typing.Optional[typing.Union[float, tuple]] = None):
        blockchain_node_urls = ([blockchain_node_url] +
                                fallback_blockchain_node_urls)
        for blockchain_node_url_ in blockchain_node_urls:
            try:
                valid_node_connection = self._create_single_node_connection(
                    blockchain_node_url_, timeout)
                if blockchain_node_url_ != blockchain_node_url:
                    fallback_blockchain_node_urls.remove(blockchain_node_url_)
                return valid_node_connection
            except SingleNodeConnectionError:
                continue
        blockchain_node_domains = [
            urllib.parse.urlparse(blockchain_node_url).netloc
            for blockchain_node_url in blockchain_node_urls
        ]
        raise self._create_error(
            'cannot connect to any of the blockchain nodes with the domains '
            f'"{blockchain_node_domains}"')

    @abc.abstractmethod
    def get_address(self, private_key: str) -> str:
        """Determine the blockchain address from a private key.

        Parameters
        ----------
        private_key : str
            The unencrypted private key.

        Returns
        -------
        str
            The blockchain address for the given private key.

        Raises
        ------
        BlockchainUtilitiesError
            If the address cannot be determined from the private key.

        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def get_balance(
            self, account_address: str,
            token_address: typing.Optional[str] = None,
            node_connections: typing.Optional[NodeConnections] = None) -> int:
        """Determine the balance of native coins or tokens of an
        address.

        Parameters
        ----------
        account_address : str
            The address that will be evaluated.
        token_address : str, optional
            The address of the token that will be interrogated.
        node_connections : NodeConnections, optional
            The node connections object to be used (default: None).

        Returns
        -------
        int
            Balance of the address.

        Raises
        ------
        BlockchainUtilitiesError
            If the balance cannot be fetched.
        ResultsNotMatchingError
            If the results given by the configured blockchain
            nodes do not match.

        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def is_valid_address(self, address: str) -> bool:
        """Determine if an address string is a valid address on the
        blockchain.

        Parameters
        ----------
        address : str
            The address string to check.

        Returns
        -------
        bool
            True if the given address string is a valid address on the
            blockchain, else False.

        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def is_equal_address(self, address_one: str, address_two: str) -> bool:
        """Determine if two addresses are equal.

        Parameters
        ----------
        address_one : str
            The first address string to check.
        address_two : str
            The second address string to check.

        Returns
        -------
        bool
            True if the given addresses are equal, else False.

        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def get_unhealthy_nodes(
            self, blockchain_nodes: list[str],
            timeout: float | tuple | None = None) -> list[UnhealthyNode]:
        """Determine the health of the blockchain nodes.

        Parameters
        ----------
        blockchain_nodes : list of str
            The URLs of the blockchain nodes to check.
        timeout : float, tuple or None
            How long to wait for the server to send data before giving up,
            as a float, or a (connect timeout, read timeout) tuple.

        Returns
        -------
        list[UnhealthyNode]
            The list of unhealthy nodes.

        """
        pass

    @abc.abstractmethod
    def _get_transaction_method_names(self) -> list[str]:
        """Determine the blockchain interactor method names which
        are sending transactions.

        Returns
        -------
        list[str]
            The list of method names which are sending
            transactions.

        """
        pass  # pragma: no cover

    def load_contract_abi(
            self,
            versioned_contract_abi: VersionedContractAbi) -> list[typing.Any]:
        """Load a contract ABI, according to its version, as a list
        from a JSON file. If a contract ABI has already been loaded
        before, a cached version is returned.

        Parameters
        ----------
        versioned_contract_abi : VersionedContractAbi
            The version and the contract ABI to load.

        Returns
        -------
        list
            The loaded contract ABI.

        """
        contract_abi = versioned_contract_abi.contract_abi
        version = versioned_contract_abi.version
        if contract_abi in self.__loaded_contract_abis:
            return self.__loaded_contract_abis[contract_abi]
        contract_abi_file_name = contract_abi.get_file_name(
            self.get_blockchain())
        versioned_contract_abi_package = (
            f'{_BASE_CONTRACT_ABI_PACKAGE}.v'
            f'{version.major}_{version.minor}_{version.patch}')
        try:
            with importlib.resources.open_text(
                    versioned_contract_abi_package,
                    contract_abi_file_name) as contract_abi_file:
                loaded_contract_abi = json.load(contract_abi_file)
            self.__loaded_contract_abis[contract_abi] = loaded_contract_abi
            return loaded_contract_abi
        except Exception:
            raise self._create_error('unable to load a contract ABI',
                                     contract_abi=contract_abi,
                                     version=version)

    @abc.abstractmethod
    def decrypt_private_key(self, encrypted_key: str, password: str) -> str:
        """Load the private key from a password-encrypted key.

        Parameters
        ----------
        encrypted_key: str
            The encrypted key.
        password : str
            The password to decrypt the key.

        Returns
        -------
        str
            The decrypted private key.

        Raises
        ------
        BlockchainUtilitiesError
            If the private key cannot be decrypted.

        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def read_transaction_status(
            self, transaction_id: str,
            node_connections: typing.Optional[NodeConnections] = None) \
            -> TransactionStatus:
        """Read the status of a transaction.

        Parameters
        ----------
        transaction_id : str
            The ID/hash of the transaction.
        node_connections : NodeConnections, optional
            The node connections object to be used (default: None).

        Returns
        -------
        TransactionStatus
            The transaction's current status.

        Raises
        ------
        BlockchainUtilitiesError
            If the transaction status cannot be read.
        ResultsNotMatchingError
            If the results given by the configured blockchain
            nodes do not match.

        """
        pass  # pragma: no cover

    @dataclasses.dataclass
    class TransactionSubmissionRequest:
        """Request data for submitting a transaction.

        Attributes
        ----------
        contract_address : BlockchainAddress
            The address of the contract to invoke a function on in the
            transaction.
        versioned_contract_abi : VersionedContractAbi
            The version and the ABI of the contract to invoke a function
            on in the transaction.
        function_selector : str
            The selector of the contract function to be invoked in the
            transaction.
        function_args : ContractFunctionArgs
            The arguments of the contract function to be invoked in the
            transaction.
        gas : int or None
            The gas to be provided for the transaction. Depending on the
            blockchain, it may not be necessary to specify the gas
            explicitly or it may be possible to estimate the required gas
            automatically.
        min_adaptable_fee_per_gas : int
            The minimum adaptable fee per gas. The definition of the
            adaptable fee per gas depends on the blockchain's
            transaction fee model.
        max_total_fee_per_gas : int or None
            The maximum total fee per gas. Since the total fee per gas
            will anyway be kept as low as possible for the transaction
            to be included in a block, it is recommended to specify a
            large maximum total fee per gas. If it is not specified at
            all, no upper limit for the total fee per gas will be
            enforced.
        amount : int or None
            The amount of native coins to be sent in the transaction
            (specified in the blockchain's smallest coin denomination).
        nonce : int
            The unique transaction nonce of the account controlled by
            the default private key.

        """
        contract_address: BlockchainAddress
        versioned_contract_abi: VersionedContractAbi
        function_selector: str
        function_args: ContractFunctionArgs
        gas: typing.Optional[int]
        min_adaptable_fee_per_gas: int
        max_total_fee_per_gas: typing.Optional[int]
        amount: typing.Optional[int]
        nonce: int

        def to_dict(self) -> dict[str, typing.Any]:
            """Convert the request instance to its corresponding
            dictionary representation.

            Returns
            -------
            dict
                The dictionary representation.

            """
            request_dict = dataclasses.asdict(self)
            request_dict['versioned_contract_abi']['contract_abi'] = \
                self.versioned_contract_abi.contract_abi.value
            request_dict['versioned_contract_abi']['version'] = \
                str(self.versioned_contract_abi.version)
            return request_dict

        @classmethod
        def from_dict(cls: type[T], request_dict: dict[str, typing.Any]) -> T:
            """Convert the dictionary representation of a request to its
            corresponding request instance.

            Parameters
            ----------
            request_dict : dict
                The dictionary representation.

            Returns
            -------
            TransactionSubmissionRequest
                The request instance.

            """
            request_dict = copy.deepcopy(request_dict)
            request_dict['versioned_contract_abi'] = VersionedContractAbi(
                ContractAbi(
                    request_dict['versioned_contract_abi']['contract_abi']),
                semantic_version.Version(
                    request_dict['versioned_contract_abi']['version']))
            return cls(**request_dict)

    @dataclasses.dataclass
    class TransactionSubmissionResponse:
        """Response data from submitting a transaction.

        Attributes
        ----------
        transaction_id : str
            The ID/hash of the submitted transaction.
        adaptable_fee_per_gas : int
            The adaptable fee per gas actually used for submitting the
            transaction.

        """
        transaction_id: str
        adaptable_fee_per_gas: int

    @abc.abstractmethod
    def submit_transaction(
            self, request: TransactionSubmissionRequest,
            node_connections: typing.Optional[NodeConnections] = None) \
            -> TransactionSubmissionResponse:
        """Submit a transaction.

        Parameters
        ----------
        request : TransactionSubmissionRequest
            The request data for submitting a transaction.
        node_connections : NodeConnections, optional
            The node connections object to be used (default: None).

        Returns
        -------
        TransactionSubmissionResponse
            The response data from submitting a transaction.

        Raises
        ------
        MaxTotalFeePerGasExceededError
            If the maximum total fee per gas would be exceeded for the
            transaction to be submitted.
        TransactionUnderpricedError
            If the transaction has been submitted as an underpriced
            transaction.
        TransactionNonceTooLowError
            If the transaction has been submitted with a nonce too low.
        ResultsNotMatchingError
            If the results given by the configured blockchain
            nodes do not match.
        BlockchainUtilitiesError
            If the transaction cannot be submitted for any other reason.

        """
        pass  # pragma: no cover

    @dataclasses.dataclass
    class TransactionResubmissionRequest(TransactionSubmissionRequest):
        """Request data for resubmitting a transaction.

        Attributes
        ----------
        adaptable_fee_increase_factor : float
            The factor for increasing the adaptable fee per gas.

        """
        adaptable_fee_increase_factor: float

    @dataclasses.dataclass
    class TransactionResubmissionResponse(TransactionSubmissionResponse):
        """Response data from resubmitting a transaction.

        """
        pass

    def resubmit_transaction(
            self, request: TransactionResubmissionRequest,
            node_connections: typing.Optional[NodeConnections] = None) \
            -> TransactionResubmissionResponse:
        """Resubmit (i.e. replace) a transaction.

        Parameters
        ----------
        request : TransactionResubmissionRequest
            The request data for resubmitting a transaction.
        node_connections : NodeConnections, optional
            The node connections object to be used (default: None).

        Returns
        -------
        TransactionResubmissionResponse
            The response data from resubmitting a transaction.

        Raises
        ------
        MaxTotalFeePerGasExceededError
            If the maximum total fee per gas would be exceeded for the
            transaction to be resubmitted.
        TransactionNonceTooLowError
            If the transaction has been resubmitted with a nonce too
            low.
        ResultsNotMatchingError
            If the results given by the configured blockchain
            nodes do not match.
        BlockchainUtilitiesError
            If the transaction cannot be resubmitted for any other
            reason.

        """
        if request.min_adaptable_fee_per_gas < 0:
            raise self._create_error(
                'previous minimum adaptable fee per gas must be >= 0',
                request=request)
        if (request.adaptable_fee_increase_factor
                < MIN_ADAPTABLE_FEE_INCREASE_FACTOR):
            raise self._create_error(
                'adaptable fee increase factor must be >= '
                f'{MIN_ADAPTABLE_FEE_INCREASE_FACTOR}', request=request)
        response = None
        while response is None:
            # Minimum adaptable fee per gas must be 1 for the
            # resubmission if the previous minimum adaptable fee per gas
            # has been 0
            min_adaptable_fee_per_gas = max(
                1,
                math.ceil(request.min_adaptable_fee_per_gas *
                          request.adaptable_fee_increase_factor))
            if (request.max_total_fee_per_gas is not None and
                    min_adaptable_fee_per_gas > request.max_total_fee_per_gas):
                raise self._create_max_total_fee_per_gas_exceeded_error(
                    request=request)
            request = dataclasses.replace(
                request, min_adaptable_fee_per_gas=min_adaptable_fee_per_gas)
            try:
                response = self.submit_transaction(request, node_connections)
            except TransactionUnderpricedError:
                _logger.warning('resubmitted transaction underpriced',
                                extra=vars(request))
        return BlockchainUtilities.TransactionResubmissionResponse(
            response.transaction_id, response.adaptable_fee_per_gas)

    @dataclasses.dataclass
    class TransactionSubmissionStartRequest(TransactionResubmissionRequest):
        """Request data for starting a transaction submission.

        Attributes
        ----------
        blocks_until_resubmission : int
            The number of blocks to wait until the transaction is
            resubmitted if it has not yet been included in a block.

        """
        blocks_until_resubmission: int

    def start_transaction_submission(
            self, request: TransactionSubmissionStartRequest,
            node_connections: typing.Optional[NodeConnections] = None) \
            -> uuid.UUID:
        """Start a transaction submission. The transaction is
        automatically resubmitted with higher transaction fees until it
        is included in a block. Celery tasks need to be enabled for this
        function.

        Parameters
        ----------
        request : TransactionSubmissionStartRequest
            The request data for starting a transaction submission.
        node_connections : NodeConnections, optional
            The node connections object to be used (default: None).

        Returns
        -------
        uuid.UUID
            The unique internal transaction ID, which can be used later
            to retrieve the status of the transaction submission.

        Raises
        ------
        MaxTotalFeePerGasExceededError
            If the maximum total fee per gas would be exceeded for the
            transaction to be submitted.
        TransactionNonceTooLowError
            If the transaction has been submitted with a nonce too low.
        ResultsNotMatchingError
            If the results given by the configured blockchain
            nodes do not match.
        BlockchainUtilitiesError
            If the transaction submission cannot be started for any
            other reason.

        """
        if not self._celery_tasks_enabled:
            raise self._create_error('Celery tasks disabled')
        if request.blocks_until_resubmission <= 0:
            raise self._create_error('blocks until resubmission must be > 0',
                                     request=request)
        if (request.adaptable_fee_increase_factor
                < MIN_ADAPTABLE_FEE_INCREASE_FACTOR):
            raise self._create_error(
                'adaptable fee increase factor must be >= '
                f'{MIN_ADAPTABLE_FEE_INCREASE_FACTOR}', request=request)
        try:
            response = self.submit_transaction(request, node_connections)
        except TransactionUnderpricedError:
            _logger.warning('initially submitted transaction underpriced',
                            extra=vars(request))
            response = self.resubmit_transaction(request, node_connections)
        try:
            from pantos.common.blockchains.tasks import \
                create_transaction_resubmission_task
            internal_transaction_id = create_transaction_resubmission_task(
                self.get_blockchain(), request, response)
        except Exception:
            raise self._create_error(
                'unable to create a transaction resubmission task',
                request=request, transaction_id=response.transaction_id)
        return internal_transaction_id

    @dataclasses.dataclass
    class TransactionSubmissionStatusResponse:
        """Response data from retrieving the status of a transaction
        submission.

        Attributes
        ----------
        transaction_submission_completed : bool
            True if and only if the transaction submission has been
            completed (i.e. the transaction is either confirmed or
            reverted).
        transaction_status : TransactionStatus or None
            The status of the submitted (and eventually included)
            transaction (available if the transaction submission has
            been completed).
        transaction_id : str or None
            The ID/hash of the submitted (and eventually included)
            transaction (available if the transaction submission has
            been completed).

        """
        transaction_submission_completed: bool
        transaction_status: typing.Optional[TransactionStatus] = None
        transaction_id: typing.Optional[str] = None

    def get_transaction_submission_status(
            self, internal_transaction_id: uuid.UUID) \
            -> TransactionSubmissionStatusResponse:
        """Retrieve the status of a transaction submission. Celery tasks
        need to be enabled for this function.

        Parameters
        ----------
        internal_transaction_id : uuid.UUID
            The unique internal transaction ID.

        Returns
        -------
        TransactionSubmissionStatusResponse
            The response data from retrieving the status of a
            transaction submission.

        Raises
        ------
        BlockchainUtilitiesError
            If the status of the transaction submission cannot be
            retrieved or if there has been an unresolvable error during
            the transaction submission.

        """
        if not self._celery_tasks_enabled:
            raise self._create_error('Celery tasks disabled')
        try:
            from pantos.common.blockchains.tasks import \
                get_transaction_resubmission_task_result
            task_result = get_transaction_resubmission_task_result(
                internal_transaction_id)
        except Exception:
            raise self._create_error(
                'unable to get a transaction resubmission task result',
                internal_transaction_id=internal_transaction_id)
        if task_result is None:
            return BlockchainUtilities.TransactionSubmissionStatusResponse(
                False)
        transaction_status = task_result[0]
        transaction_id = task_result[1]
        return BlockchainUtilities.TransactionSubmissionStatusResponse(
            True, transaction_status, transaction_id)

    @abc.abstractmethod
    def _create_single_node_connection(
            self, blockchain_node_url: str,
            timeout: float | tuple | None = None) -> typing.Any:
        """Create a single blockchain-specific node connection
        with the given URL.

        Parameters
        ----------
        blockchain_node_url : str
            The blockchain node URL.
        timeout : float, tuple or None
            How long to wait for the server to send data before giving up,
            as a float, or a (connect timeout, read timeout) tuple.

        Returns
        -------
        typing.Any
            The blockchain-specific node connection.

        Raises
        ------
        SingleNodeConnectionError
            If the node connection cannot be established.

        """
        pass  # pragma: no cover

    def _create_max_total_fee_per_gas_exceeded_error(
            self, **kwargs: typing.Any) -> BlockchainUtilitiesError:
        return self._create_error(
            specialized_error_class=MaxTotalFeePerGasExceededError, **kwargs)

    def _create_transaction_nonce_too_low_error(
            self, **kwargs: typing.Any) -> BlockchainUtilitiesError:
        return self._create_error(
            specialized_error_class=TransactionNonceTooLowError, **kwargs)

    def _create_transaction_underpriced_error(
            self, **kwargs: typing.Any) -> BlockchainUtilitiesError:
        return self._create_error(
            specialized_error_class=TransactionUnderpricedError, **kwargs)

    def _create_single_node_connection_error(
            self, **kwargs: typing.Any) -> BlockchainUtilitiesError:
        return self._create_error(
            specialized_error_class=SingleNodeConnectionError, **kwargs)



import getpass
import typing

import Crypto.PublicKey.ECC
import Crypto.Signature.eddsa


class SignerError(Exception):
    """Exception class for signer errors.

    """
    def __init__(self, message: str, name: str = 'signer error',
                 **kwargs: typing.Any):
        """Construct a signer error.

        Parameters
        ----------
        message : str
            Explanation of the error.
        name : str
            Human-readable name of the error.
        **kwargs : dict
            Additional information about the error as keyword arguments.

        """
        super().__init__(message, name, **kwargs)


class _Signer:
    def __init__(self, pem_value: str, pem_password: str):
        """
        Constructor of Signer class.

        Parameters
        ----------
        pem_value : str
            Value of the encrypted private key.
        pem_password : str
            Password to unlock the PEM file.

        """
        self.__signer = self._load_signer(pem_value, pem_password)

    def sign_message(self, message: str) -> str:
        """Sign a message.

        Parameters
        ----------
        message : str
            The message to be signed.

        Returns
        -------
        str
            The signature of the message.

        Raises
        ------
        SignerError
            If the message cannot be signed.

        """
        try:
            message_bytes = message.encode()
            signature = self.__signer.sign(message_bytes)
            return signature.hex()
        except Exception:
            raise SignerError(
                f'unable to compute signature of message: {message}')

    def verify_message(self, message: str, signature: str) -> bool:
        """Verify that the message is valid (signed by the same
        private key).

        Parameters
        ----------
        message : str
            The message to be verified.
        signature : str
            The signature of the message.

        Returns
        -------
        bool
            If the signature of the message is valid.

        Raises
        ------
        SignerError
            If the message cannot be verified.

        """
        try:
            message_bytes = message.encode()
            try:
                self.__signer.verify(message_bytes, bytes.fromhex(signature))
                return True
            except ValueError:
                return False
        except Exception:
            raise SignerError(
                f'unable to verify signature of message {message}')

    def build_message(self, separator: str = '', *message_parts:
                      typing.Any) -> str:
        """Build a message.

        Parameters
        ----------
        message_parts : typing.Any
            Ordered number of message parts to be concatenated into one
            message.
        separator : str
            String which is separating each part of the message.

        Returns
        -------
        str
            The built message.

        """
        message = ''
        for message_part in message_parts:
            message += f'{str(message_part)}{separator}'
        # cutting of last separator
        return message[:-1]

    def _load_signer(
            self, pem_value: str,
            pem_password: str) -> Crypto.Signature.eddsa.EdDSASigScheme:
        """Load the EdDSA signer object from a password-encrypted pem file.
        The key must be on the curve Ed25519 or Ed448.

        Parameters
        ----------
        pem_value : str
            Value of the encrypted private key.
        pem_password : str
            Password to unlock the PEM file.

        Returns
        -------
        Crypto.Signature.eddsa.EdDSASigScheme
            An EdDSA signature object.

        Raises
        ------
        SignerError
            If the EdDSA signer cannot be loaded.

        """
        try:
            if pem_password is None:
                pem_password = getpass.getpass(
                    'Password for decrypting the pem file')

            private_key = Crypto.PublicKey.ECC.import_key(
                pem_value, passphrase=pem_password)
            return Crypto.Signature.eddsa.new(private_key,
                                              'rfc8032')  # type: ignore
        except SignerError:
            raise
        except Exception:
            raise SignerError('cannot load the private key')


_signer: typing.Optional[_Signer] = None


def get_signer(pem_value: str, pem_password: str) -> _Signer:
    """Get a _Signer object.

    Parameters
    ----------
    pem_value : str
        Value of the encrypted private key.
    pem_password : str
        Password to unlock the PEM file.

    Returns
    -------
    _Signer
        Signer object used for signing and verifying messages.

    Raises
    ------
    SignerError
        If the signer cannot be gotten.

    """
    global _signer
    if not _signer:
        _signer = _Signer(pem_value, pem_password)
    return _signer



import dataclasses
import logging
import typing
import uuid

import celery  # type: ignore
import celery.result  # type: ignore

from pantos.common.blockchains.base import BlockchainUtilities
from pantos.common.blockchains.base import MaxTotalFeePerGasExceededError
from pantos.common.blockchains.enums import Blockchain
from pantos.common.blockchains.factory import get_blockchain_utilities
from pantos.common.entities import TransactionStatus

_MAX_TRANSACTION_RESUBMISSION_TASK_RETRIES = 1000

TransactionResubmissionRequest = \
    BlockchainUtilities.TransactionResubmissionRequest

TransactionSubmissionResponse = \
    BlockchainUtilities.TransactionSubmissionResponse

TransactionSubmissionStartRequest = \
    BlockchainUtilities.TransactionSubmissionStartRequest

assert (sorted(
    field.name
    for field in dataclasses.fields(TransactionResubmissionRequest)) == sorted(
        field.name
        for field in dataclasses.fields(TransactionSubmissionStartRequest)
        if field.name != 'blocks_until_resubmission'))

_logger = logging.getLogger(__name__)


def create_transaction_resubmission_task(
        blockchain: Blockchain, request: TransactionSubmissionStartRequest,
        response: TransactionSubmissionResponse) -> uuid.UUID:
    """Create a new transaction resubmission task execution after the
    initial transaction submission.

    Parameters
    ----------
    blockchain : Blockchain
        The blockchain the transaction has been submitted to.
    request : TransactionSubmissionStartRequest
        The original request for starting the transaction submission.
    response : TransactionSubmissionResponse
        The response from the initial transaction submission.

    Returns
    -------
    uuid.UUID
        The internal transaction ID associated with the task.

    Raises
    ------
    Exception
        If Celery has an issue with submitting the new task.

    """
    internal_transaction_id = uuid.uuid4()
    blockchain_utilities = get_blockchain_utilities(blockchain)
    request = dataclasses.replace(
        request, min_adaptable_fee_per_gas=response.adaptable_fee_per_gas)
    request_dict = request.to_dict()
    del request_dict['blocks_until_resubmission']
    task_args = (blockchain.value, request.blocks_until_resubmission,
                 response.transaction_id, request_dict)
    task_id = str(internal_transaction_id)
    countdown = (blockchain_utilities.average_block_time *
                 request.blocks_until_resubmission)
    _transaction_resubmission_task.apply_async(args=task_args, task_id=task_id,
                                               countdown=countdown)
    return internal_transaction_id


def get_transaction_resubmission_task_result(
        internal_transaction_id: uuid.UUID) \
        -> typing.Optional[tuple[TransactionStatus, str]]:
    """Get the result of a transaction resubmission task execution.

    Parameters
    ----------
    internal_transaction_id : uuid.UUID
        The internal transaction ID associated with the task.

    Returns
    -------
    tuple or None
        None if the task has not yet finished, else a pair of the status
        and ID of the submitted transaction.

    Raises
    ------
    Exception
        If there has been an unresolvable error during the task
        execution.

    """
    task_id = str(internal_transaction_id)
    task_result = celery.result.AsyncResult(task_id)
    if not task_result.ready():
        return None
    assert task_result.state in ['SUCCESS', 'FAILURE']
    transaction_status_id, transaction_id = task_result.get(
        disable_sync_subtasks=False)
    return TransactionStatus(transaction_status_id), transaction_id


@celery.shared_task(bind=True,
                    max_retries=_MAX_TRANSACTION_RESUBMISSION_TASK_RETRIES)
def _transaction_resubmission_task(
        self, blockchain_id: int, blocks_until_resubmission: int,
        transaction_id: str,
        request_dict: dict[str, typing.Any]) -> tuple[int, str]:
    blockchain = Blockchain(blockchain_id)
    blockchain_utilities = get_blockchain_utilities(blockchain)
    task_info = request_dict | {
        'blockchain': blockchain,
        'blocks_until_resubmission': blocks_until_resubmission,
        'transaction_id': transaction_id,
        'internal_transaction_id': self.request.id
    }
    resubmission_countdown = (blockchain_utilities.average_block_time *
                              blocks_until_resubmission)
    confirmation_countdown = (
        blockchain_utilities.average_block_time *
        blockchain_utilities.required_transaction_confirmations)
    try:
        transaction_status = blockchain_utilities.read_transaction_status(
            transaction_id)
    except Exception as error:
        _logger.error('unable to read the transaction status', extra=task_info,
                      exc_info=True)
        raise self.retry(countdown=resubmission_countdown, exc=error)
    _logger.info(f'transaction {transaction_status.name.lower()}',
                 extra=task_info)
    if transaction_status is TransactionStatus.UNINCLUDED:
        request = TransactionResubmissionRequest.from_dict(request_dict)
        try:
            response = blockchain_utilities.resubmit_transaction(request)
        except MaxTotalFeePerGasExceededError as error:
            _logger.warning(
                'unable to further increase the adaptable fee per gas',
                extra=task_info)
            countdown = confirmation_countdown * self.request.retries
            raise self.retry(countdown=countdown, exc=error)
        except Exception as error:
            _logger.error('unable to resubmit a transaction', extra=task_info,
                          exc_info=True)
            raise self.retry(countdown=resubmission_countdown, exc=error)
        _logger.info(
            'adaptable fee per gas increased to '
            f'{response.adaptable_fee_per_gas}', extra=task_info)
        request.min_adaptable_fee_per_gas = response.adaptable_fee_per_gas
        task_args = (blockchain_id, blocks_until_resubmission,
                     response.transaction_id, request.to_dict())
        raise self.retry(args=task_args, countdown=resubmission_countdown)
    if transaction_status is TransactionStatus.UNCONFIRMED:
        raise self.retry(countdown=confirmation_countdown)
    assert transaction_status in [
        TransactionStatus.CONFIRMED, TransactionStatus.REVERTED
    ]
    return transaction_status.value, transaction_id



"""Module for analyzing the health of the system.

"""
import concurrent.futures
import dataclasses

from pantos.common.blockchains.base import UnhealthyNode
from pantos.common.blockchains.enums import Blockchain
from pantos.common.blockchains.factory import get_blockchain_utilities
from pantos.common.exceptions import NotInitializedError

_blockchain_nodes: dict[Blockchain, tuple[list[str],
                                          float | tuple | None]] = {}


@dataclasses.dataclass
class NodesHealth:
    """Entity which provides information about the health status
    of the nodes requested for a blockchain network.

    Attributes
    ----------
    healthy_total : int
        The total number of healthy nodes.
    unhealthy_total : int
        The total number of unhealthy nodes.
    unhealthy_nodes : list[str, str]
        The list of unhealthy nodes with their respective status.

    """
    healthy_total: int
    unhealthy_total: int
    unhealthy_nodes: list[UnhealthyNode]


def check_blockchain_nodes_health() -> dict[Blockchain, NodesHealth]:
    """Check the health of the blockchain nodes.

    Returns
    -------
    dict[Blockchain, NodesHealth]
        The health status of the blockchain nodes.

    Raises
    ------
    NotInitializedError
        If the blockchain nodes have not been initialized yet.

    """
    if _blockchain_nodes == {}:
        raise NotInitializedError(
            'the blockchain nodes have not been initialized yet')
    nodes_health = {}

    with concurrent.futures.ThreadPoolExecutor() as executor:
        future_to_blockchain = {
            executor.submit(
                get_blockchain_utilities(blockchain).get_unhealthy_nodes,  # noqa
                _blockchain_nodes[blockchain][0],
                _blockchain_nodes[blockchain][1]): blockchain
            for blockchain in _blockchain_nodes
        }
        for future in concurrent.futures.as_completed(future_to_blockchain):
            blockchain = future_to_blockchain[future]
            unhealthy_nodes = future.result()
            nodes_health[blockchain] = NodesHealth(
                len(_blockchain_nodes[blockchain][0]) - len(unhealthy_nodes),
                len(unhealthy_nodes), unhealthy_nodes)
    return nodes_health


def initialize_blockchain_nodes(
    blockchain_nodes: dict[Blockchain, tuple[list[str],
                                             float | tuple | None]]) \
        -> None:
    """Initialize the blockchain nodes.

    Parameters
    ----------
    blockchain_nodes : dict[Blockchain, tuple[list[str], float | tuple]]
        The blockchain nodes to be initialized.

    """
    global _blockchain_nodes
    if _blockchain_nodes != blockchain_nodes:  # pragma: no cover
        _blockchain_nodes = blockchain_nodes



"""Module for communicating with Pantos service nodes.

"""
import dataclasses
import typing
import uuid

import requests

from pantos.common.blockchains.enums import Blockchain
from pantos.common.entities import ServiceNodeBid
from pantos.common.entities import ServiceNodeTransferStatus
from pantos.common.exceptions import BaseError
from pantos.common.types import BlockchainAddress

_TRANSFER_RESOURCE = 'transfer'
_STATUS_RESOURCE = 'status'
_BID_RESOURCE = 'bids'


class ServiceNodeClientError(BaseError):
    """Exception class for all service node client errors.

    """
    pass


class ServiceNodeClient:
    """Client for communicating with Pantos service nodes.

    """
    @dataclasses.dataclass
    class SubmitTransferRequest:
        """Request data for submitting a new token transfer request to a
        Pantos service node.

        Attributes
        ----------
        service_node_url : str
            The chosen service node's base URL.
        source_blockchain: Blockchain
            The token transfer's source blockchain.
        destination_blockchain: Blockchain
            The token transfer's destination blockchain.
        sender_address: BlockchainAddress
            The address of the sender on the source blockchain.
        recipient_address: BlockchainAddress
            The address of the recipient on the destination blockchain.
        source_token_address: BlockchainAddress
            The address of the transferred token on the source
            blockchain.
        destination_token_address: BlockchainAddress
            The address of the transferred token on the destination
            blockchain.
        token_amount: int
            The amount of tokens to be transferred.
        service_node_bid: ServiceNodeBid
            The chosen service node bid.
        sender_nonce: int
            The unique sender nonce for the new token transfer.
        valid_until: int
            The timestamp until when the token transfer can be included
            on the source blockchain (in seconds since the epoch).
        signature: str
            The sender's signature for the new token transfer.

        """
        service_node_url: str
        source_blockchain: Blockchain
        destination_blockchain: Blockchain
        sender_address: BlockchainAddress
        recipient_address: BlockchainAddress
        source_token_address: BlockchainAddress
        destination_token_address: BlockchainAddress
        token_amount: int
        service_node_bid: ServiceNodeBid
        sender_nonce: int
        valid_until: int
        signature: str

    @dataclasses.dataclass
    class TransferStatusResponse:
        """Response data for checking the status of a transfer at a
        service node.

        Attributes
        ----------
        task_id: uuid.UUID
            The unique task ID of the token transfer.
        source_blockchain: Blockchain
            The token transfer's source blockchain.
        destination_blockchain: Blockchain
            The token transfer's destination blockchain.
        sender_address: BlockchainAddress
            The address of the sender on the source blockchain.
        recipient_address: BlockchainAddress
            The address of the recipient on the destination blockchain.
        source_token_address: BlockchainAddress
            The address of the transferred token on the source
            blockchain.
        destination_token_address: BlockchainAddress
            The address of the transferred token on the destination
            blockchain.
        token_amount: int
            The amount of tokens transferred.
        fee: int
            The fee paid to the service node for the transfer.
        status: ServiceNodeTransferStatus
            The service node transfer status.
        transfer_id: int
            The Pantos transfer ID.
        transaction_id: str
            The ID/hash of the token transfer's transaction.

        """
        task_id: uuid.UUID
        source_blockchain: Blockchain
        destination_blockchain: Blockchain
        sender_address: BlockchainAddress
        recipient_address: BlockchainAddress
        source_token_address: BlockchainAddress
        destination_token_address: BlockchainAddress
        token_amount: int
        fee: int
        status: ServiceNodeTransferStatus
        transfer_id: int
        transaction_id: str

    def submit_transfer(self, request: SubmitTransferRequest,
                        timeout: typing.Optional[float] = None) -> uuid.UUID:
        """Submit a new token transfer request to a Pantos service node.

        Parameters
        ----------
        request : SubmitTransferRequest
            The request data for a new token transfer.

        Returns
        -------
        uuid.UUID
            The service node's task ID.

        Raises
        ------
        ServiceNodeClientError
            If the token transfer request cannot be submitted
            successfully.

        """
        service_node_request = {
            'source_blockchain_id': request.source_blockchain.value,
            'destination_blockchain_id': request.destination_blockchain.value,
            'sender_address': request.sender_address,
            'recipient_address': request.recipient_address,
            'source_token_address': request.source_token_address,
            'destination_token_address': request.destination_token_address,
            'amount': request.token_amount,
            'bid': {
                'fee': request.service_node_bid.fee,
                'execution_time': request.service_node_bid.execution_time,
                'valid_until': request.service_node_bid.valid_until,
                'signature': request.service_node_bid.signature
            },
            'nonce': request.sender_nonce,
            'valid_until': request.valid_until,
            'signature': request.signature
        }
        transfer_url = self.__build_transfer_url(request.service_node_url)
        try:
            service_node_response = requests.post(transfer_url,
                                                  json=service_node_request,
                                                  timeout=timeout)
            # Raise an error in case of a 4xx or 5xx response status code
            service_node_response.raise_for_status()
            task_id = service_node_response.json()['task_id']
            return uuid.UUID(task_id)
        except (requests.exceptions.RequestException, KeyError):
            response_message = self.__read_response_message(
                service_node_response)
            raise ServiceNodeClientError(
                'unable to submit a new token transfer request',
                request=request, transfer_url=transfer_url,
                response_message=response_message)

    def bids(
            self, service_node_url: str, source_blockchain: Blockchain,
            destination_blockchain: Blockchain,
            timeout: typing.Optional[float] = None) \
            -> typing.List[ServiceNodeBid]:
        """Retrieve the bids of the service node found at the given
        service node url.

        Parameters
        ----------
        service_node_url : str
            The url of the service node.
        source_blockchain : Blockchain
            The source blockchain of the bid.
        destination_blockchain : Blockchain
            The destination blockchain of the bid.

        Returns
        -------
        list of ServiceNodeBid
            The list of service node bid given by the service node.

        Raises
        -------
        ServiceNodeClientError
            If unable to retrieve the bids.

        """
        bids_url = self.__build_bids_url(service_node_url,
                                         str(source_blockchain.value),
                                         str(destination_blockchain.value))
        try:
            service_node_response = requests.get(bids_url, timeout=timeout)
            service_node_response.raise_for_status()
            bids = service_node_response.json()
            response = []
            for bid in bids:
                response.append(
                    ServiceNodeBid(source_blockchain, destination_blockchain,
                                   bid['fee'], bid['execution_time'],
                                   bid['valid_until'], bid['signature']))
            return response
        except (requests.exceptions.RequestException, ValueError, KeyError):
            response_message = self.__read_response_message(
                service_node_response)
            raise ServiceNodeClientError(
                'unable to get the bids of the service node',
                service_node_url=service_node_url,
                source_blockchain=source_blockchain,
                destination_blockchain=destination_blockchain,
                response_message=response_message)

    def status(
            self, service_node_url: str, task_id: uuid.UUID,
            timeout: typing.Optional[float] = None) -> TransferStatusResponse:
        """Retrieve the status of a transfer.

        Parameters
        ----------
        service_node_url : str
            The url of the service node.
        task_id : uuid.UUID
            The task id of the transfer.

        Returns
        -------
        TransferStatusResponse
            The transfer status response.

        Raises
        ------
        ServiceNodeClientError
            If unable to get the status of a transfer.

        """
        status_url = self.__build_status_url(service_node_url, task_id)
        try:
            service_node_response = requests.get(status_url, timeout=timeout)
            service_node_response.raise_for_status()
            json_response = service_node_response.json()
            transfer_status_response = self.TransferStatusResponse(
                uuid.UUID(json_response['task_id']),
                Blockchain(json_response['source_blockchain_id']),
                Blockchain(json_response['destination_blockchain_id']),
                BlockchainAddress(json_response['sender_address']),
                BlockchainAddress(json_response['recipient_address']),
                BlockchainAddress(json_response['source_token_address']),
                BlockchainAddress(json_response['destination_token_address']),
                json_response['amount'], json_response['fee'],
                ServiceNodeTransferStatus.from_name(json_response['status']),
                json_response['transfer_id'], json_response['transaction_id'])
            return transfer_status_response
        except (requests.exceptions.RequestException, ValueError, KeyError):
            response_message = self.__read_response_message(
                service_node_response)
            raise ServiceNodeClientError(
                'unable to get the status of the transfer',
                service_node_url=service_node_url, task_id=task_id,
                response_message=response_message)

    def __build_transfer_url(self, service_node_url: str) -> str:
        transfer_url = service_node_url
        if not service_node_url.endswith('/'):
            transfer_url += '/'
        transfer_url += _TRANSFER_RESOURCE
        return transfer_url

    def __build_bids_url(self, service_node_url: str, source_blockchain: str,
                         destination_blockchain: str) -> str:
        bids_url = service_node_url
        if not service_node_url.endswith('/'):
            bids_url += '/'
        return (f'{bids_url}{_BID_RESOURCE}?'
                f'source_blockchain={source_blockchain}&'
                f'destination_blockchain={destination_blockchain}')

    def __build_status_url(self, service_node_url: str,
                           task_id: uuid.UUID) -> str:
        transfer_url = self.__build_transfer_url(service_node_url)
        return f'{transfer_url}/{str(task_id)}/{_STATUS_RESOURCE}'

    def __read_response_message(
            self, response: requests.Response) -> typing.Optional[str]:
        response_message = None
        if 'application/json' in response.headers.get('content-type', ''):
            response_message = response.json().get('message')
        return response_message



"""Module for Ethereum-specific utilities and errors.

"""
import logging
import typing
import urllib.parse

import semantic_version  # type: ignore
import web3
import web3.contract.contract
import web3.exceptions
import web3.middleware
import web3.types

from pantos.common.blockchains.base import GENERAL_RPC_ERROR_MESSAGE
from pantos.common.blockchains.base import BlockchainUtilities
from pantos.common.blockchains.base import BlockchainUtilitiesError
from pantos.common.blockchains.base import NodeConnections
from pantos.common.blockchains.base import ResultsNotMatchingError
from pantos.common.blockchains.base import SingleNodeConnectionError
from pantos.common.blockchains.base import UnhealthyNode
from pantos.common.blockchains.base import VersionedContractAbi
from pantos.common.blockchains.enums import Blockchain
from pantos.common.blockchains.enums import ContractAbi
from pantos.common.entities import TransactionStatus
from pantos.common.types import BlockchainAddress

_NONCE_TOO_LOW = ['nonce too low', 'invalid nonce', 'ERR_INCORRECT_NONCE']
"""Possible nonce too low error messages."""

_TRANSACTION_METHOD_NAMES = [
    'send_transaction', 'replace_transaction', 'modify_transaction',
    'send_raw_transaction', 'transact'
]
"""The names of methods of the blockchain interactor object which
send transactions."""

_NO_ARCHIVE_NODE_RPC_ERROR_MESSAGE = 'missing trie node'

_NO_ARCHIVE_NODE_LOG_MESSAGE = 'due to the absence of an archive node'

_logger = logging.getLogger(__name__)


class EthereumUtilitiesError(BlockchainUtilitiesError):
    """Exception class for all Ethereum utilities errors.

    """
    pass


class EthereumUtilities(BlockchainUtilities):
    """Class for Ethereum-specific utilities.

    """
    def __init__(self, blockchain_node_urls: list[str],
                 fallback_blockchain_node_urls: list[str],
                 average_block_time: int,
                 required_transaction_confirmations: int,
                 transaction_network_id: typing.Optional[int],
                 default_private_key: typing.Optional[tuple[str, str]] = None,
                 celery_tasks_enabled: bool = False):
        # Docstring inherited
        if transaction_network_id is None:
            raise self._create_error(
                'transaction network ID (chain ID) must be given')
        super().__init__(blockchain_node_urls, fallback_blockchain_node_urls,
                         average_block_time,
                         required_transaction_confirmations,
                         transaction_network_id,
                         default_private_key=default_private_key,
                         celery_tasks_enabled=celery_tasks_enabled)

    def create_node_connections(
            self,
            timeout: typing.Optional[typing.Union[float, tuple]] = None) \
            -> NodeConnections[web3.Web3]:
        # Docstring inherited
        return super().create_node_connections(timeout)

    def create_contract(
            self, contract_address: BlockchainAddress,
            versioned_contract_abi: VersionedContractAbi,
            node_connections: typing.Optional[NodeConnections] = None) \
            -> NodeConnections.Wrapper[web3.contract.Contract]:
        """Create a contract instance.

        Parameters
        ----------
        contract_address : BlockchainAddress
            The address of the contract.
        versioned_contract_abi : VersionedContractAbi
            The version and the contract ABI to load.
        w3 : web3.Web3, optional
            The Web3 instance to use.

        Returns
        -------
        NodeConnections.Wrapper[web3.contract.Contract]
            The wrapper instance over the contract object.

        Raises
        ------
        EthereumUtilitiesError
            If the contract instance cannot be created.

        """
        if not self.is_valid_address(contract_address):
            raise self._create_error('invalid contract address',
                                     contract_address=contract_address)
        node_connections = self.__get_node_connections(node_connections)
        try:
            return node_connections.eth.contract(
                address=typing.cast(web3.types.ChecksumAddress,
                                    contract_address),
                abi=self.load_contract_abi(versioned_contract_abi))
        except Exception:
            raise self._create_error(
                'unable to create a contract instance',
                contract_address=contract_address,
                contract_abi=versioned_contract_abi.contract_abi,
                version=versioned_contract_abi.version)

    def get_address(self, private_key: str) -> str:
        # Docstring inherited
        try:
            address = web3.Account.from_key(private_key).address
            assert address.startswith('0x')
            return address
        except Exception:
            raise self._create_error(
                'cannot determine the address from a private key')

    def get_balance(
            self, account_address: str,
            token_address: typing.Optional[str] = None,
            node_connections: typing.Optional[NodeConnections] = None) -> int:
        # Docstring inherited
        if not self.is_valid_address(account_address):
            raise self._create_error('invalid account address')
        node_connections = self.__get_node_connections(node_connections)
        if token_address is None:
            try:
                return node_connections.eth.get_balance(
                    typing.cast(web3.types.ChecksumAddress,
                                account_address)).get()
            except ResultsNotMatchingError:
                raise
            except Exception:
                raise self._create_error('cannot determine balance')
        else:
            versioned_contract_abi = VersionedContractAbi(
                ContractAbi.STANDARD_TOKEN, semantic_version.Version('1.0.0'))
            if not self.is_valid_address(token_address):
                raise self._create_error('invalid token address')
            erc20_contract = self.create_contract(
                BlockchainAddress(token_address), versioned_contract_abi,
                node_connections)
            try:
                return erc20_contract.functions.\
                    balanceOf(account_address).call().get()
            except ResultsNotMatchingError:
                raise
            except Exception:
                raise self._create_error("cannot call the contract")

    @classmethod
    def get_blockchain(cls) -> Blockchain:
        # Docstring inherited
        return Blockchain.ETHEREUM

    @classmethod
    def get_error_class(cls) -> type[BlockchainUtilitiesError]:
        # Docstring inherited
        return EthereumUtilitiesError

    def get_logs(
            self,
            event: NodeConnections.Wrapper[
                web3.contract.contract.ContractEvent],
            from_block_number: int, to_block_number: int) \
            -> typing.List[web3.types.EventData]:
        """Retrieves the logs of a contract event from a block range.

        Parameters
        ----------
        event : NodeConnections.Wrapper[
                    web3.contract.contract.ContractEvent]
            The instance of the wrapper over the contract event,
            used for retrieving the associated logs.
        from_block_number : int
            The block number to start the search from.
        to_block_number : int
            The block number to end the search at.

        Returns
        -------
        list of web3.types.EventData
            The logs of the contract event.

        Raises
        ------
        ResultsNotMatchingError
            If the results given by the configured blockchain
            nodes do not match.
        BlockchainUtilitiesError
            If the balance cannot be fetched.

        """
        try:
            # Query all events of the contract event.address between the two
            # specified block numbers
            logs = event.get_logs(fromBlock=from_block_number,
                                  toBlock=to_block_number, argument_filters={
                                      'address': event.address
                                  }).get()
            return logs
        except ResultsNotMatchingError:
            raise
        except Exception:
            assert event is not None
            raise self._create_error(
                'cannot process the "{}" event logs of the contract with the '
                'address "{}"'.format(event.event_name, event.address))

    def is_valid_address(self, address: str) -> bool:
        # Docstring inherited
        return web3.Web3.is_checksum_address(address)

    def is_equal_address(self, address_one: str, address_two: str) -> bool:
        # Docstring inherited
        return address_one.lower() == address_two.lower()

    def get_unhealthy_nodes(
            self, blockchain_nodes: list[str],
            timeout: float | tuple | None = None) -> list[UnhealthyNode]:
        # Docstring inherited
        unhealthy_nodes = []
        for blockchain_node in blockchain_nodes:
            try:
                self._create_single_node_connection(blockchain_node, timeout)
            except SingleNodeConnectionError:
                unhealthy_nodes.append(
                    UnhealthyNode(
                        urllib.parse.urlparse(blockchain_node).netloc,
                        GENERAL_RPC_ERROR_MESSAGE))
        return unhealthy_nodes

    def _get_transaction_method_names(self) -> list[str]:
        # Docstring inherited
        return _TRANSACTION_METHOD_NAMES

    def decrypt_private_key(self, encrypted_key: str, password: str) -> str:
        # Docstring inherited
        try:
            private_key = web3.Account.decrypt(encrypted_key, password).hex()
        except Exception:
            raise self._create_error('cannot load the private key')
        assert not private_key.startswith('0x')
        return private_key

    def read_transaction_status(
            self, transaction_id: str,
            node_connections: typing.Optional[NodeConnections] = None) \
            -> TransactionStatus:
        # Docstring inherited
        try:
            node_connections = self.__get_node_connections(node_connections)
            try:
                transaction_receipt = \
                    node_connections.eth.get_transaction_receipt(
                        typing.cast(web3.types.HexStr, transaction_id)).get()
            except web3.exceptions.TransactionNotFound:
                return TransactionStatus.UNINCLUDED
            assert (transaction_receipt['transactionHash'].to_0x_hex() ==
                    transaction_id)
            transaction_block_number = transaction_receipt['blockNumber']
            if transaction_block_number is None:
                return TransactionStatus.UNINCLUDED
            current_block_number = \
                node_connections.eth.get_block_number().get_minimum_result()
            confirmations = current_block_number - transaction_block_number
            assert confirmations >= 0
            if confirmations < self.required_transaction_confirmations:
                return TransactionStatus.UNCONFIRMED
            transaction_status = transaction_receipt['status']
            if transaction_status != 1:
                revert_message = self.__retrieve_revert_message(
                    transaction_id, node_connections)
                _logger.info(revert_message, extra=transaction_receipt)
                return TransactionStatus.REVERTED
            return TransactionStatus.CONFIRMED
        except ResultsNotMatchingError:
            raise
        except Exception:
            raise self._create_error(
                'unable to read the status of a transaction',
                transaction_id=transaction_id)

    def submit_transaction(
            self, request: BlockchainUtilities.TransactionSubmissionRequest,
            node_connections: typing.Optional[NodeConnections] = None) \
            -> BlockchainUtilities.TransactionSubmissionResponse:
        # Docstring inherited
        if self._default_private_key is None:
            raise self._create_error('default private key must be available')
        self.__check_transaction_submission_request(request)
        try:
            node_connections = self.__get_node_connections(node_connections)
            transaction_parameters, adaptable_fee_per_gas = \
                self.__create_transaction_parameters(request, node_connections)
            _logger.info('new transaction to be submitted',
                         extra=vars(request) | transaction_parameters)
            contract = self.create_contract(request.contract_address,
                                            request.versioned_contract_abi,
                                            node_connections)
            contract_function = contract.get_function_by_selector(
                request.function_selector)
            transaction = contract_function(
                *request.function_args).build_transaction(
                    transaction_parameters).get()
            signed_transaction = web3.Account.sign_transaction(
                transaction, private_key=self._default_private_key)
            transaction_hash = self.__send_raw_transaction(
                signed_transaction.rawTransaction, node_connections)
            return BlockchainUtilities.TransactionSubmissionResponse(
                transaction_hash, adaptable_fee_per_gas)
        except EthereumUtilitiesError:
            raise
        except ResultsNotMatchingError:
            raise
        except Exception:
            raise self._create_error('unable to submit a transaction',
                                     request=request)

    def _create_single_node_connection(
            self, blockchain_node_url: str,
            timeout: float | tuple | None = None) -> typing.Any:
        # Docstring inherited
        request_kwargs = {'timeout': timeout}
        try:
            w3 = web3.Web3(
                web3.Web3.HTTPProvider(blockchain_node_url,
                                       request_kwargs=request_kwargs))
            if w3.is_connected():
                try:
                    w3.eth.get_block('latest')
                except web3.exceptions.ExtraDataLengthError:
                    w3.middleware_onion.inject(
                        web3.middleware.geth_poa_middleware, layer=0)
                _logger.info(
                    'new blockchain node connection', extra={
                        'blockchain': self.get_blockchain(),
                        'blockchain_node_domain': urllib.parse.urlparse(
                            blockchain_node_url).netloc,
                        'client_version': w3.client_version
                    })
                return w3
        except Exception:
            raise self._create_single_node_connection_error()
        raise self._create_single_node_connection_error()

    def _type_2_transactions_supported(self) -> bool:
        return True  # pragma: no cover

    def __check_transaction_submission_request(
            self, request: BlockchainUtilities.TransactionSubmissionRequest) \
            -> None:
        if request.gas is not None and request.gas < 21000:
            raise self._create_error('gas must be >= 21000', request=request)
        if request.min_adaptable_fee_per_gas < 0:
            raise self._create_error(
                'minimum adaptable fee per gas must be >= 0', request=request)
        if (request.max_total_fee_per_gas is not None
                and request.min_adaptable_fee_per_gas
                > request.max_total_fee_per_gas):
            raise self._create_max_total_fee_per_gas_exceeded_error(
                request=request)
        if request.amount is not None and request.amount < 0:
            raise self._create_error('amount must be >= 0', request=request)
        if request.nonce < 0:
            raise self._create_error('nonce must be >= 0', request=request)

    def __create_transaction_parameters(
            self, request: BlockchainUtilities.TransactionSubmissionRequest,
            node_connections: NodeConnections[web3.Web3]) -> \
            tuple[web3.types.TxParams, int]:
        assert self.transaction_network_id is not None
        assert self._default_address is not None
        transaction_parameters: web3.types.TxParams = {
            'chainId': self.transaction_network_id,
            'from': self._default_address,
            'nonce': web3.types.Nonce(request.nonce)
        }
        if request.gas is not None:
            # If gas is not explicitly specified, it is automatically
            # estimated using w3.eth.estimate_gas()
            transaction_parameters['gas'] = web3.types.Wei(request.gas)
        if request.amount is not None:
            transaction_parameters['value'] = web3.types.Wei(request.amount)
        if self._type_2_transactions_supported():
            # EIP-1559 transaction
            base_fee_per_gas = node_connections.eth.get_block(
                'latest')['baseFeePerGas'].get_minimum_result()
            max_priority_fee_per_gas = request.min_adaptable_fee_per_gas
            max_fee_per_gas = 2 * base_fee_per_gas + max_priority_fee_per_gas
            if (request.max_total_fee_per_gas is not None
                    and request.max_total_fee_per_gas < max_fee_per_gas):
                _logger.warning(
                    f'maximum total fee per gas < {max_fee_per_gas}',
                    extra=vars(request))
                max_fee_per_gas = request.max_total_fee_per_gas
            transaction_parameters['maxPriorityFeePerGas'] = web3.types.Wei(
                max_priority_fee_per_gas)
            transaction_parameters['maxFeePerGas'] = web3.types.Wei(
                max_fee_per_gas)
            adaptable_fee_per_gas = max_priority_fee_per_gas
        else:
            # Legacy type-0 transaction
            gas_price = max(
                node_connections.eth.gas_price.get_minimum_result(),
                request.min_adaptable_fee_per_gas)
            if (request.max_total_fee_per_gas is not None
                    and request.max_total_fee_per_gas < gas_price):
                _logger.warning(f'maximum total fee per gas < {gas_price}',
                                extra=vars(request))
                gas_price = request.max_total_fee_per_gas
            transaction_parameters['gasPrice'] = web3.types.Wei(gas_price)
            adaptable_fee_per_gas = gas_price
        return transaction_parameters, adaptable_fee_per_gas

    def __send_raw_transaction(
            self, raw_transaction: bytes,
            node_connections: NodeConnections[web3.Web3]) -> str:
        try:
            return typing.cast(
                str,
                node_connections.eth.send_raw_transaction(
                    raw_transaction).to_0x_hex())
        except ValueError as error:
            if any(error_message in str(error)
                   for error_message in _NONCE_TOO_LOW):
                raise self._create_transaction_nonce_too_low_error()
            if 'transaction underpriced' in str(error):
                raise self._create_transaction_underpriced_error()
            raise

    def __get_node_connections(
            self, node_connections: typing.Optional[NodeConnections] = None) \
            -> NodeConnections[web3.Web3]:
        if node_connections is None:
            node_connections = self.create_node_connections()
        return node_connections

    def __retrieve_revert_message(
            self, transaction_hash: str,
            node_connections: NodeConnections[web3.Web3]) -> str:
        revert_message = 'unknown'
        try:
            full_tx = node_connections.eth.get_transaction(
                typing.cast(web3.types.HexStr, transaction_hash)).get()
            replay_tx = {
                'from': full_tx['from'],
                'to': full_tx['to'],
                'value': full_tx['value'],
                'data': full_tx['input']
            }
            context_block_number = full_tx['blockNumber'] - 1
            try:
                node_connections.eth.call(
                    typing.cast(web3.types.TxParams, replay_tx),
                    context_block_number).get()
            except web3.exceptions.ContractLogicError as error:
                revert_message = str(error)
            except ValueError as error:
                if _NO_ARCHIVE_NODE_RPC_ERROR_MESSAGE in error.args[0].get(
                        'message'):
                    revert_message += f' {_NO_ARCHIVE_NODE_LOG_MESSAGE}'
        except Exception:
            _logger.warning('unable to retrieve the revert message',
                            exc_info=True)
        return revert_message



"""Module for initializing console and file logging.

"""
import dataclasses
import datetime
import enum
import logging
import logging.handlers
import pathlib
import sys
import traceback
import typing

import json_log_formatter  # type: ignore

from pantos.common.blockchains.enums import Blockchain

_HUMAN_READABLE_LOG_FORMAT: typing.Final[str] = (
    '%(asctime)s - %(name)s - %(thread)d - %(levelname)s - '
    '%(message)s%(extra)s')


@dataclasses.dataclass
class LogFile:
    """Properties of a log file.

    Attributes
    ----------
    file_path : pathlib.Path
        The path to the log file.
    max_bytes : int
        Maximum number of bytes the log file can reach before it is
        rolled over.
    backup_count : int
        Number of old log files kept (by appending the file extensions
        .1, .2, etc.).

    """
    file_path: pathlib.Path
    max_bytes: int
    backup_count: int


class LogFormat(enum.Enum):
    """Enumeration of available log formats.

    """
    HUMAN_READABLE = 1
    JSON = 2

    @staticmethod
    def from_name(name: str) -> 'LogFormat':
        """Find an enumeration member by its name.

        Parameters
        ----------
        name : str
            The name to search for.

        Raises
        ------
        NameError
            If no enumeration member can be found for the given name.

        """
        name_upper = name.upper()
        for log_format in LogFormat:
            if name_upper == log_format.name:
                return log_format
        raise NameError(name)


class _DataDogJSONFormatter(json_log_formatter.VerboseJSONFormatter):
    """Custom JSON formatter tailored for proper DataDog logs.

    """
    def mutate_json_record(self, json_record: typing.Dict[str | int,
                                                          typing.Any]):
        """Convert fields of `json_record` to needed types.

        Parameters
        ----------
        json_record : dict of str or int, typing.Any
            The json dictionary object of the log.

        Returns
        -------
        dict of str or int, typing.Any
            The mutated dictionary object of the log.

        """
        for attribute_name in json_record:
            attribute = json_record[attribute_name]
            if isinstance(attribute, datetime.datetime):
                json_record[attribute_name] = attribute.isoformat()
            if isinstance(attribute, Blockchain):
                json_record[attribute_name] = attribute.name_in_pascal_case
        return json_record

    def json_record(self, message: str, extra: typing.Dict[str | int,
                                                           typing.Any],
                    record: typing.Any) -> typing.Dict[str | int, typing.Any]:
        """Prepares a JSON payload which will be logged.

        Parameters
        ----------
        message : str
            Log message.
        extra : dict of str or int, typing.Any
            Dictionary that was passed as `extra` parameter.
        record : typing.Any
            Log record returned by JSONFormatter.format()

        Returns
        ------
        dict of str or int, typing.Any
            The dictionary of the record.

        """
        extra['levelname'] = record.levelname
        if 'time' not in extra:
            extra['time'] = datetime.datetime.utcnow()

        if record.exc_info:
            exc_type = record.exc_info[0]
            exc_value = record.exc_info[1]
            human_readable_error = traceback.format_exception_only(
                exc_type, exc_value)[-1].rstrip()
            extra['message'] = f'{message} - {human_readable_error}'
            extra['exc_info'] = self.formatException(record.exc_info)
        else:
            extra['message'] = message

        return extra


class _HumanReadableFormatter(logging.Formatter):
    """Formatter for human-readable log messages.

    """
    def __init__(self):
        log_record = logging.LogRecord(None, None, None, None, None, None,
                                       None)
        self.__log_record_attributes = set(log_record.__dict__.keys())
        self.__log_record_attributes.add('asctime')
        self.__log_record_attributes.add('message')
        super().__init__(_HUMAN_READABLE_LOG_FORMAT)

    def format(self, log_record: logging.LogRecord) -> str:
        # Docstring inherited
        extra = ''
        for key, value in log_record.__dict__.items():
            if key not in self.__log_record_attributes and key != 'extra':
                extra += f' - {key}: {value}'
        log_record.__dict__['extra'] = extra
        return super().format(log_record)


def initialize_logger(logger: logging.Logger = logging.getLogger(),
                      log_format: LogFormat = LogFormat.HUMAN_READABLE,
                      standard_output: bool = True,
                      log_file: typing.Optional[LogFile] = None,
                      debug: bool = False) -> None:
    """Initialize a logger by setting its log format, output streams,
    and the level of logged messages.

    Parameters
    ----------
    logger : logging.Logger
        The logger to be initialized (default: root logger).
    log_format : LogFormat
        The format of the log output (default: human-readable format).
    standard_output : bool
        If True, the messages are logged to the standard output, which
        is shown on the console if it is not redirected (default: True).
    log_file : LogFile, optional
        If given, messages are logged to the specified log file
        (default: None).
    debug : bool
        If True, debug messages are logged (default: False).

    Raises
    ------
    OSError
        If the logs cannot be written to the specified log file.

    """
    logger.handlers.clear()
    formatter = _create_formatter(log_format)
    if standard_output:
        logger.addHandler(_create_standard_output_handler(formatter))
    if log_file is not None:
        logger.addHandler(_create_rotating_file_handler(log_file, formatter))
    logger.setLevel(logging.DEBUG if debug else logging.INFO)


def _create_formatter(log_format: LogFormat) -> logging.Formatter:
    if log_format is LogFormat.HUMAN_READABLE:
        return _HumanReadableFormatter()
    if log_format is LogFormat.JSON:
        return _DataDogJSONFormatter()
    raise NotImplementedError


def _create_rotating_file_handler(
        log_file: LogFile, formatter: logging.Formatter) -> logging.Handler:
    if not log_file.file_path.parent.exists():
        log_file.file_path.parent.mkdir(parents=True)
    handler = logging.handlers.RotatingFileHandler(
        log_file.file_path, maxBytes=log_file.max_bytes,
        backupCount=log_file.backup_count)
    handler.setFormatter(formatter)
    return handler


def _create_standard_output_handler(
        formatter: logging.Formatter) -> logging.Handler:
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)
    return handler



"""Module that defines Pantos entities.

"""
import dataclasses
import enum
import typing

from pantos.common.blockchains.enums import Blockchain
from pantos.common.types import Amount
from pantos.common.types import BlockchainAddress


@dataclasses.dataclass
class ServiceNodeBid:
    """Entity that represents a Pantos service node bid.

    Attributes
    ----------
    source_blockchain : Blockchain
        The source blockchain of token transfers covered by the
        service node bid.
    destination_blockchain : Blockchain
        The destination blockchain of token transfers covered by the
        service node bid.
    fee : Amount
        The service node's fee for a token transfer.
    execution_time : int
        The time (in seconds since the epoch) the service node has
        available for executing a token transfer on the source
        blockchain.
    valid_until : int
        The time (in seconds since the epoch) until which the bid is
        valid.
    signature : str
        The signature over the the service node's bid. The signature is over
        ```
        sig(fee,
             bid_valid_until,
             source_blockchain_id,
             destination_blockchain_id,
             execution_time)
        ```

    """
    source_blockchain: Blockchain
    destination_blockchain: Blockchain
    fee: Amount
    execution_time: int
    valid_until: int
    signature: str


@dataclasses.dataclass
class TokenDeploymentRequest:
    """Request data for submitting a new deployment request to the
    Pantos token creator service.

    Attributes
    ----------
    deployment_blockchain_ids : list of int
        The deployment blockchain ids.
    token_name : str
        The name of the token.
    token_symbol : str
        The symbol of the token.
    token_decimals : int
        The number of decimals of the token.
    token_pausable : bool
        If the token is pausable.
    token_burnable : bool
        If the token is burnable.
    token_supply : int
        The supply of the token.
    payment_blockchain_id : int
        The blockchain id of the blockchain where the deployment
        fee will be paid.
    payer_address : BlockchainAddress
        The blockchain address of the payer.
    deployment_fee : int
        The deployment fee to be paid.
    deployment_fee_signature : str
        The deployment fee signature.
    deployment_fee_valid_until : int
        The timestamp until when the deployment fee can be submitted
        to the token creator service (in seconds since the epoch).
    bid_fee : Amount
        The fee of the bid.
    bid_execution_time : int
        The execution time of the bid.
    bid_valid_until : int
        The time (in seconds since the epoch) until the
        service node is able to execute the token
        transfer on the source blockchain.
    bid_signature : str
        The signature of the bid given by the service node.
    payment_nonce : int
        The nonce of the payment.
    payment_valid_until : int
        The timestamp until when the payment can be submitted
        to the service node (in seconds since the epoch).
    payment_signature : str
        The signature of the payment.

    """
    deployment_blockchain_ids: typing.List[int]
    token_name: str
    token_symbol: str
    token_decimals: int
    token_pausable: bool
    token_burnable: bool
    token_supply: int
    payment_blockchain_id: int
    payer_address: BlockchainAddress
    deployment_fee: int
    deployment_fee_signature: str
    deployment_fee_valid_until: int
    bid_fee: Amount
    bid_execution_time: int
    bid_valid_until: int
    bid_signature: str
    payment_nonce: int
    payment_valid_until: int
    payment_signature: str


BlockchainAddressBidPair = typing.Tuple[BlockchainAddress, ServiceNodeBid]


class ServiceNodeTransferStatus(enum.IntEnum):
    """Enumeration of possible transfer status values.

    """
    ACCEPTED = 0
    FAILED = 1
    SUBMITTED = 2
    REVERTED = 3
    CONFIRMED = 4

    @staticmethod
    def from_name(name: str) -> 'ServiceNodeTransferStatus':
        """Convert a string to a ServiceNodeTransferStatus object
        (if possible).

        Parameters
        ----------
        name : str
            The name of the status to be converted.

        Raises
        ------
        NameError
            If the type conversion is not possible.

        """
        for status in ServiceNodeTransferStatus:
            if name.upper() == status.name:
                return status
        raise NameError(name)


class TransactionStatus(enum.Enum):
    """Enumeration of blockchain transaction status values.

    """
    UNINCLUDED = 0
    UNCONFIRMED = 1
    CONFIRMED = 2
    REVERTED = 3



"""Common REST API resources.

"""
import dataclasses
import json
import logging

import flask  # type: ignore
import flask_restful  # type: ignore

from pantos.common.exceptions import NotInitializedError
from pantos.common.health import check_blockchain_nodes_health

_logger = logging.getLogger(__name__)
"""Logger for this module."""


class Live(flask_restful.Resource):
    """Flask resource class which specifies the health/live REST endpoint.

    """
    def get(self):
        """Return null response with status 200 if the service is up
        and running.

        """
        return None  # pragma: no cover


class NodesHealthResource(flask_restful.Resource):
    """RESTful resource for the health status of the blockchain nodes.

    """
    def get(self):
        """Return the health status of the blockchain nodes.

        """
        try:
            _logger.info('checking blockchain nodes health')
            nodes_health = check_blockchain_nodes_health()
            return ok_response({
                blockchain.name.capitalize(): dataclasses.asdict(
                    nodes_health[blockchain])
                for blockchain in nodes_health
            })
        except NotInitializedError:
            _logger.warning('no blockchain nodes have been initialized yet')
            return internal_server_error(
                'no blockchain nodes have been initialized yet')
        except Exception:
            _logger.critical('cannot check blockchain nodes health',
                             exc_info=True)
            return internal_server_error()


def ok_response(data: list | dict) -> flask.Response:
    """Create a Flask response given some data.

    Parameters
    ----------
    data : list or dict
        The data that will be wrapped in a Flask response.

    Returns
    -------
    flask.Response
        The Flask response object.

    """
    return flask.Response(json.dumps(data), status=200,
                          mimetype='application/json')


def no_content_response() -> flask.Response:
    """Create a Flask response for a successful request without any
    content.

    Returns
    -------
    flask.Response
        The Flask response object.

    """
    return flask.Response(status=204)


def conflict(error_message: str):
    """Raise an HTTPException when a request conflicts
    with the current state of the server.

    Parameters
    ----------
    error_message : str
        The error message.

    Raises
    ------
    HTTPException
        HTTP exception raised with the code 409.

    """
    flask_restful.abort(409, message=error_message)


def not_acceptable(error_messages: str | list | dict):
    """Raise an HTTPException for non-acceptable requests.

    Parameters
    ----------
    error_messages : str or list or dict
        The error messages.

    Raises
    ------
    HTTPException
        HTTP exception raised with the code 406.

    """
    flask_restful.abort(406, message=error_messages)


def bad_request(error_messages: str | list | dict):
    """Raise an HTTPException for bad requests.

    Parameters
    ----------
    error_messages : str or list or dict
        The error messages.

    Raises
    ------
    HTTPException
        HTTP exception raised with the code 400.

    """
    flask_restful.abort(400, message=error_messages)


def forbidden(error_message: str):
    """Raise an HTTPException if a prohibited action is refused.

    Parameters
    ----------
    error_message : str
        The error message.

    Raises
    ------
    HTTPException
        HTTP exception raised with the code 403.

    """
    flask_restful.abort(403, message=error_message)


def resource_not_found(error_message: str):
    """Raise an HTTPException if the resource has not been found.

    Parameters
    ----------
    error_message : str
        The error message.

    Raises
    ------
    HTTPException
        HTTP exception raised with the code 404.

    """
    flask_restful.abort(404, message=error_message)


def internal_server_error(error_message: str | None = None):
    """Raise an HTTPException for internal server errors.

    Parameters
    ----------
    error_message : str, optional
        The error message.

    Raises
    ------
    HTTPException
        HTTP exception raised with the code 500.

    """
    flask_restful.abort(500, message=error_message)



"""Module for Solana-specific utilities and errors.

"""
import typing

from pantos.common.blockchains.base import BlockchainUtilities
from pantos.common.blockchains.base import BlockchainUtilitiesError
from pantos.common.blockchains.base import NodeConnections
from pantos.common.blockchains.base import UnhealthyNode
from pantos.common.blockchains.enums import Blockchain
from pantos.common.entities import TransactionStatus


class SolanaUtilitiesError(BlockchainUtilitiesError):
    """Exception class for all Solana utilities errors.

    """
    pass


class SolanaUtilities(BlockchainUtilities):
    """Class for Solana-specific utilities.

    """
    def __init__(self, blockchain_node_urls: list[str],
                 fallback_blockchain_node_urls: list[str],
                 average_block_time: int,
                 required_transaction_confirmations: int,
                 transaction_network_id: typing.Optional[int],
                 default_private_key: typing.Optional[tuple[str, str]] = None,
                 celery_tasks_enabled: bool = False):
        # Docstring inherited
        super().__init__(blockchain_node_urls, fallback_blockchain_node_urls,
                         average_block_time,
                         required_transaction_confirmations,
                         transaction_network_id,
                         default_private_key=default_private_key,
                         celery_tasks_enabled=celery_tasks_enabled)

    def get_address(self, private_key: str) -> str:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def get_balance(
            self, account_address: str,
            token_address: typing.Optional[str] = None,
            node_connections: typing.Optional[NodeConnections] = None) -> int:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def get_blockchain(cls) -> Blockchain:
        # Docstring inherited
        return Blockchain.SOLANA

    @classmethod
    def get_error_class(cls) -> type[BlockchainUtilitiesError]:
        # Docstring inherited
        return SolanaUtilitiesError

    def is_valid_address(self, address: str) -> bool:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def is_equal_address(self, address_one: str, address_two: str) -> bool:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def get_unhealthy_nodes(
            self, blockchain_nodes: list[str],
            timeout: float | tuple | None = None) -> list[UnhealthyNode]:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def _get_transaction_method_names(self) -> list[str]:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def decrypt_private_key(self, encrypted_key: str, password: str) -> str:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def read_transaction_status(
            self, transaction_id: str,
            node_connections: typing.Optional[NodeConnections] = None) \
            -> TransactionStatus:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def submit_transaction(
            self, request: BlockchainUtilities.TransactionSubmissionRequest,
            node_connections: typing.Optional[NodeConnections] = None) \
            -> BlockchainUtilities.TransactionSubmissionResponse:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def _create_single_node_connection(
            self, blockchain_node_url: str,
            timeout: typing.Optional[typing.Union[float,
                                                  tuple]] = None) \
            -> typing.Any:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover



"""Module for Cronos-specific utilities and errors. Since Cronos is
Ethereum-compatible, the utilities implementation inherits from the
pantos.common.blockchains.ethereum module.

"""
from pantos.common.blockchains.base import BlockchainUtilitiesError
from pantos.common.blockchains.enums import Blockchain
from pantos.common.blockchains.ethereum import EthereumUtilities
from pantos.common.blockchains.ethereum import EthereumUtilitiesError


class CronosUtilitiesError(EthereumUtilitiesError):
    """Exception class for all Cronos utilities errors.

    """
    pass


class CronosUtilities(EthereumUtilities):
    """Class for Cronos-specific utilities.

    """
    @classmethod
    def get_blockchain(cls) -> Blockchain:
        # Docstring inherited
        return Blockchain.CRONOS

    @classmethod
    def get_error_class(cls) -> type[BlockchainUtilitiesError]:
        # Docstring inherited
        return CronosUtilitiesError



"""Module that defines value object types.

"""
import collections.abc
import decimal
import typing


class BlockchainAddress(str):
    pass


class PrivateKey(str):
    pass


class TokenSymbol(str):
    pass


AccountId = typing.Union[BlockchainAddress, PrivateKey]

Amount = typing.Union[int, decimal.Decimal]

ContractFunctionArgs = collections.abc.Sequence[typing.Union[
    bool, int, str, collections.abc.Sequence[typing.Union[bool, int, str]]]]

TokenId = typing.Union[BlockchainAddress, TokenSymbol]
